from typing import List, Type
import requests
import pandas as pd
import warnings
import os
import shutil


from ITR.configs import PortfolioCoverageTVPConfig
from ITR.interfaces import IDataProviderCompany


class SBTi:
    """
    Data provider skeleton for SBTi. This class only provides the sbti_validated field for existing companies.
    """

    def _check_if_cta_file_exists(self):
        """
        Check if the CTA file exists in the local file system
        """
        return os.path.isfile(self.c.FILE_TARGETS)

    def _get_bundled_cta_path(self):
        """
        Get the path to the bundled CTA file in the package
        """
        return os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "..",
            "inputs",
            "current-Companies-Taking-Action.xlsx",
        )

    def _use_bundled_fallback(self):
        """
        Copy the bundled CTA file to the target location as a fallback
        """
        bundled_path = self._get_bundled_cta_path()
        if os.path.isfile(bundled_path):
            try:
                target_dir = os.path.dirname(self.c.FILE_TARGETS)
                os.makedirs(target_dir, exist_ok=True)
                shutil.copy2(bundled_path, self.c.FILE_TARGETS)
                print(f'WARNING: Using bundled CTA file as fallback. Data may be outdated.')
                print(f'Copied bundled file to: {self.c.FILE_TARGETS}')
                return True
            except Exception as e:
                print(f'Error copying bundled CTA file: {e}')
                return False
        else:
            print(f'ERROR: Bundled CTA file not found at {bundled_path}')
            return False


    def _check_CTA_less_than_one_week_old(self):
        """
        Check if the CTA file is older than a week
        """
        file_updated_time = os.path.getmtime(self.c.FILE_TARGETS)
        week_in_seconds = 7 * 24 * 60 * 60
        # If the file was updated more than a week ago, return False
        if file_updated_time < pd.Timestamp.now().timestamp() - week_in_seconds:
            return False
        else:
            return True

    def _use_local_cta_file(self):
        if self.c.FILE_TARGETS_CUSTOM_PATH is None:
            raise ValueError('Please set FILE_TARGETS_CUSTOM_PATH to the path of the CTA file.')
        self.c.FILE_TARGETS = self.c.FILE_TARGETS_CUSTOM_PATH

        # check that file is not more than a week old
        if  self._check_if_cta_file_exists():
            # file_age = os.path.getmtime(self.c.FILE_TARGETS)
            # week_in_seconds = 7 * 24 * 60 * 60 # frequency of CTA file updates

            if not self._check_CTA_less_than_one_week_old(): 
                print(f'CTA file is older than a week, if you want to keep your file up-to-date please update the file at {self.c.FILE_TARGETS}.')
        else:
            raise ValueError('CTA file does not exist')

    def _download_cta_file(self):
        get_file = False
        if self._check_if_cta_file_exists():
            if self.c.SKIP_CTA_FILE_IF_EXISTS:
                if not self._check_CTA_less_than_one_week_old():
                    get_file = True
            else:
                get_file = True
        else:                  
            get_file = True

        if get_file:
            download_successful = False
            try:
                self._fetch_and_save_cta_file()
                download_successful = True
            
            except requests.HTTPError as err:
                if err.response.status_code == 403:
                    print(f'403 Error fetching the CTA file: {err}')
                else:
                    print(f'HTTP Error fetching the CTA file: {err}')
            except requests.RequestException as err:
                print(f'Network error fetching the CTA file: {err}')
                print(f'This may be due to firewall restrictions or network connectivity issues.')
            except Exception as err:
                print(f'Unexpected error fetching the CTA file: {err}')
            
            # If download failed and file still doesn't exist, use bundled fallback
            if not download_successful and not self._check_if_cta_file_exists():
                print(f'Attempting to use bundled CTA file as fallback...')
                if not self._use_bundled_fallback():
                    raise RuntimeError(
                        'Failed to download CTA file and no fallback available. '
                        'Please check network connectivity or set USE_LOCAL_CTA=True '
                        'with a valid FILE_TARGETS_CUSTOM_PATH.'
                    )
        else:            
            print(f'CTA file already exists in {self.c.FILE_TARGETS}, skipping download.')

    def _fetch_and_save_cta_file(self):
        # Ensure the directory exists
        target_dir = os.path.dirname(self.c.FILE_TARGETS)
        os.makedirs(target_dir, exist_ok=True)
        
        headers = {
            'x-request': 'download',
            'User-Agent': 'ITR-tool/0.9.2 (Python; ekonomi-finans@wwf.se)',
            'From': 'ekonomi-finans@wwf.se'
        }
        # read from the remote CTA file url
        response = requests.get(self.c.CTA_FILE_URL, headers=headers, timeout=30)
        # raise if the status code is not 200
        response.raise_for_status()

        with open(self.c.FILE_TARGETS, 'wb') as output:
            output.write(response.content)
            print(f'Successfully downloaded CTA file (Status: {response.status_code})')
            print(f'Saved to: {self.c.FILE_TARGETS}')

    def handle_cta_file(self):
        if self.c.USE_LOCAL_CTA:
            self._use_local_cta_file()
        else:
            self._download_cta_file()

    def __init__(
        self, config: Type[PortfolioCoverageTVPConfig] = PortfolioCoverageTVPConfig
    ):
        self.c = config

        # Handle CTA file
        self.handle_cta_file()

        # Read CTA file into pandas dataframe
        # Suppress warning about openpyxl - check if this is still needed in the released version.
        warnings.filterwarnings('ignore', category=UserWarning, module='openpyxl')
       
        path = os.path.realpath(self.c.FILE_TARGETS)
        self.targets = pd.read_excel(path)

        
    def filter_cta_file(self, targets):
        """
        Filter the CTA file to create a dataframe that has one row per company
        with the columns "Action" and "Target".
        If Action = Target then only keep the rows where Target = Near-term.
        """

        # Create a new dataframe with only the columns "Action" and "Target"
        # and the columns that are needed for identifying the company
        targets = targets[
            [
                self.c.COL_COMPANY_NAME, 
                self.c.COL_COMPANY_ISIN, 
                self.c.COL_COMPANY_LEI, 
                self.c.COL_TARGET
            ]
        ]
        
        # Keep rows where Action = Target and Target = Near-term 
        df_nt_targets = targets[
           (targets[self.c.COL_TARGET] == self.c.VALUE_STATUS_SET)]
        
        # Drop duplicates in the dataframe by waterfall. 
        # Do company name last due to risk of misspelled names
        # First drop duplicates on LEI, then on ISIN, then on company name
        df_nt_targets = pd.concat([
            df_nt_targets[~df_nt_targets[self.c.COL_COMPANY_LEI].isnull()].drop_duplicates(
                subset=self.c.COL_COMPANY_LEI, keep='first'
            ), 
            df_nt_targets[df_nt_targets[self.c.COL_COMPANY_LEI].isnull()]
        ])
        
        df_nt_targets = pd.concat([
            df_nt_targets[~df_nt_targets[self.c.COL_COMPANY_ISIN].isnull()].drop_duplicates(
                subset=self.c.COL_COMPANY_ISIN, keep='first'
            ),
            df_nt_targets[df_nt_targets[self.c.COL_COMPANY_ISIN].isnull()]
        ])

        df_nt_targets.drop_duplicates(subset=self.c.COL_COMPANY_NAME, inplace=True)
  
        return df_nt_targets
    
    def get_sbti_targets(
        self, companies: List[IDataProviderCompany], id_map: dict
    ) -> List[IDataProviderCompany]:
        """
        Check for each company if they have an SBTi validated target, first using the company LEI, 
        if available, and then using the ISIN.
        
        :param companies: A list of IDataProviderCompany instances
        :param id_map: A map from company id to a tuple of (ISIN, LEI)
        :return: A list of IDataProviderCompany instances, supplemented with the SBTi information
        """
        # Filter out information about targets
        self.targets = self.filter_cta_file(self.targets)

        for company in companies:
            id_tuple = id_map.get(company.company_id)
            if id_tuple is None:
                continue
            isin, lei = id_tuple
            # Check lei and length of lei to avoid zeros
            # Handle both None and string values
            if lei is not None and str(lei).lower() not in ('nan', 'none', '') and len(str(lei)) > 3:
                targets = self.targets[
                    self.targets[self.c.COL_COMPANY_LEI] == lei
                ]
            elif isin is not None and str(isin).lower() not in ('nan', 'none', ''):
                targets = self.targets[
                    self.targets[self.c.COL_COMPANY_ISIN] == isin
                ]
            else:
                continue
            if len(targets) > 0:
                company.sbti_validated = (
                    self.c.VALUE_STATUS_SET in targets[self.c.COL_TARGET].values
                )
        return companies

   


"""
This module contains classes that create connections to data providers.
"""

from .data_provider import DataProvider
from .csv import CSVProvider
from .excel import ExcelProvider
from .bloomberg import Bloomberg
from .cdp import CDP
from .iss import ISS
from .trucost import Trucost
from .urgentem import Urgentem





import logging
import pandas as pd
import ast
from typing import List, Optional, Tuple, Type, Dict

from ITR.configs import ColumnsConfig
from ITR.data.sbti import SBTi
from ITR.interfaces import IDataProviderTarget, IDataProviderCompany


from .interfaces import PortfolioCompany, EScope, ETimeFrames, ScoreAggregations
from .target_validation import TargetProtocol

from .temperature_score import Scenario, TemperatureScore
from .portfolio_aggregation import PortfolioAggregationMethod

from . import data


DATA_PROVIDER_MAP: Dict[str, Type[data.DataProvider]] = {
    "excel": data.ExcelProvider,
    "csv": data.CSVProvider,
    "bloomberg": data.Bloomberg,
    "cdp": data.CDP,
    "iss": data.ISS,
    "trucost": data.Trucost,
    "urgentem": data.Urgentem,
}


def get_data_providers(
    data_providers_configs: List[dict], data_providers_input: List[str]
) -> List[data.DataProvider]:
    """
    Determines which data provider and in which order should be used.

    :param data_providers_configs: A list of data provider configurations
    :param data_providers_input: A list of data provider names
    :return: a list of data providers in order.
    """
    #we need to see if we can remove this method so we first throw an error if the method is called
    raise NotImplementedError("This method is not implemented yet")

    logger = logging.getLogger(__name__)
    data_providers = []
    for data_provider_config in data_providers_configs:
        data_provider_config["class"] = DATA_PROVIDER_MAP[data_provider_config["type"]](
            **data_provider_config["parameters"]
        )
        data_providers.append(data_provider_config)

    selected_data_providers = []
    for data_provider_name in data_providers_input:
        found = False
        for data_provider_config in data_providers:
            if data_provider_config["name"] == data_provider_name:
                selected_data_providers.append(data_provider_config["class"])
                found = True
                break
        if not found:
            logger.warning(
                "The following data provider could not be found: {}".format(
                    data_provider_name
                )
            )

    if len(selected_data_providers) == 0:
        raise ValueError(
            "None of the selected data providers are available. The following data providers are valid "
            "options: "
            + ", ".join(
                data_provider["name"] for data_provider in data_providers_configs
            )
        )
    return selected_data_providers


def get_company_data(
    data_providers: list, company_ids: List[str]
) -> List[IDataProviderCompany]:
    """
    Get the company data in a waterfall method, given a list of companies and a list of data providers. This will go
    through the list of data providers and retrieve the required info until either there are no companies left or there
    are no data providers left.

    :param data_providers: A list of data providers instances
    :param company_ids: A list of company ids (ISINs)
    :return: A data frame containing the company data
    """
    company_data = []
    logger = logging.getLogger(__name__)
    for dp in data_providers:
        try:
            company_data_provider = dp.get_company_data(company_ids)
            company_data += company_data_provider
            company_ids = [
                company
                for company in company_ids
                if company not in [c.company_id for c in company_data_provider]
            ]
            if len(company_ids) == 0:
                break
        except NotImplementedError:
            logger.warning("{} is not available yet".format(type(dp).__name__))

    return company_data


def get_targets(data_providers: list, companies: list) -> List[IDataProviderTarget]:
    """
    Get the targets in a waterfall method, given a list of companies and a list of data providers. This will go through
    the list of data providers and retrieve the required info until either there are no companies left or there are no
    data providers left.

    :param data_providers: A list of data providers instances
    :param companies: A list of companies. Each company should be a dict and contain a company_name and company_id field
    :return: A data frame containing the targets
    """
    target_data = []
    logger = logging.getLogger(__name__)
    for dp in data_providers:
        try:
            targets_data_provider = dp.get_targets(companies)
            target_data += targets_data_provider
            companies = [
                company
                for company in companies
                if company not in [t.company_id for t in targets_data_provider]
            ]
            if len(companies) == 0:
                break
        except NotImplementedError:
            logger.warning("{} is not available yet".format(type(dp).__name__))

    return target_data


def _flatten_user_fields(record: PortfolioCompany):
    """
    Flatten the user fields in a portfolio company and return it as a dictionary.

    :param record: The record to flatten
    :return:
    """
    record_dict = record.model_dump(exclude_none=True)
    if record.user_fields is not None:
        for key, value in record_dict["user_fields"].items():
            record_dict[key] = value
        del record_dict["user_fields"]

    return record_dict

def _make_id_map(df_portfolio: pd.DataFrame) -> dict:
    """
    Create a mapping from company_id to ISIN and LEI (required for the SBTi matching).

    :param df_portfolio: The complete portfolio
    :return: A mapping from company_id to (ISIN, LEI) tuple
    """
    # Ensure ISIN and LEI columns exist (may be absent when all values are None
    # due to exclude_none=True in model_dump)
    for col in [ColumnsConfig.COMPANY_ISIN, ColumnsConfig.COMPANY_LEI]:
        if col not in df_portfolio.columns:
            df_portfolio[col] = None

    return {
        company_id: (company[ColumnsConfig.COMPANY_ISIN], company[ColumnsConfig.COMPANY_LEI])
        for company_id, company in df_portfolio[
            [ColumnsConfig.COMPANY_ID, ColumnsConfig.COMPANY_ISIN, ColumnsConfig.COMPANY_LEI]
        ]
        .set_index(ColumnsConfig.COMPANY_ID)
        .to_dict(orient="index")
        .items()
    }


def dataframe_to_portfolio(df_portfolio: pd.DataFrame) -> List[PortfolioCompany]:
    """
    Convert a data frame to a list of portfolio company objects.

    :param df_portfolio: The data frame to parse. The column names should align with the attribute names of the
    PortfolioCompany model.
    :return: A list of portfolio companies
    """
    # df_portfolio[ColumnsConfig.ENGAGEMENT_TARGET] = (
    #     df_portfolio[ColumnsConfig.ENGAGEMENT_TARGET].fillna(False).astype("bool")
    # )
    # Set the future behavior option
    pd.set_option('future.no_silent_downcasting', True)

    df_portfolio[ColumnsConfig.ENGAGEMENT_TARGET] = (
        df_portfolio[ColumnsConfig.ENGAGEMENT_TARGET].fillna(False).infer_objects().astype("bool")
    )
    
    if 'user_fields' in df_portfolio:
        def process_user_fields(x):
            if isinstance(x, str):
                x = x.strip()
                if x.startswith('{') or x.startswith('['):
                    try:
                        return ast.literal_eval(x)
                    except (ValueError, SyntaxError):
                        pass
                return {'category': x}
            return None

        df_portfolio['user_fields'] = df_portfolio['user_fields'].apply(process_user_fields)
    return [
        PortfolioCompany.model_validate(company)
        for company in df_portfolio.to_dict(orient="records")
    ]

def check_sbti_validation(
    targets: List[IDataProviderTarget], 
    companies: List[IDataProviderCompany]) -> List[IDataProviderTarget]:
    """
    Loops through targets and sets the value of sbti_validated to True
    if the associated company's status of sbti_validated is True, otherwise False
    We need this information to correctly use the method's time frame exception.
    This column is dropped before merging with company data in the process method
    of target_validation.

    :param targets: A list of IDataProviderTarget instances
    :param companies: A list of IDataProviderCompany instances
    """
    # Create a mapping from company_id to sbti_validated status
    company_sbti_map = {c.company_id: getattr(c, "sbti_validated", False) for c in companies}
    for t in targets:
        t.sbti_validated = company_sbti_map.get(t.company_id, False)
    return targets


def get_data(
    data_providers: List[data.DataProvider], portfolio: List[PortfolioCompany]
) -> pd.DataFrame:
    """
    Get the required data from the data provider(s), validate the targets and return a 9-box grid for each company.

    :param data_providers: A list of DataProvider instances
    :param portfolio: A list of PortfolioCompany models
    :return: A data frame containing the relevant company-target data
    """
    df_portfolio = pd.DataFrame.from_records(
        [_flatten_user_fields(c) for c in portfolio]
    )
    company_data = get_company_data(data_providers, df_portfolio["company_id"].tolist())
    target_data = get_targets(data_providers, df_portfolio["company_id"].tolist())

    if len(target_data) == 0:
        raise ValueError("No targets found")

    # Supplement the company and target data with the SBTi target status
    company_data = SBTi().get_sbti_targets(company_data, _make_id_map(df_portfolio))
    target_data = check_sbti_validation(target_data, company_data)
    
    # Prepare the data
    portfolio_data = TargetProtocol().process(target_data, company_data)
    portfolio_data = pd.merge(
        left=portfolio_data,
        right=df_portfolio.drop("company_name", axis=1),
        how="left",
        on=["company_id"],
    )

    if len(company_data) == 0:
        raise ValueError(
            "None of the companies in your portfolio could be found by the data providers"
        )

    return portfolio_data


def calculate(
    portfolio_data: pd.DataFrame,
    default_score: float,
    aggregation_method: PortfolioAggregationMethod,
    grouping: Optional[List[str]],
    scenario: Optional[Scenario],
    time_frames: List[ETimeFrames],
    scopes: List[EScope],
    anonymize: bool,
    aggregate: bool = True,
) -> Tuple[pd.DataFrame, Optional[ScoreAggregations]]:
    """
    Calculate the different parts of the temperature score (actual scores, aggregations, column distribution).

    :param portfolio_data: The portfolio data, already processed by the target validation module
    :param default_score: The default score to use while calculating the temperature score
    :param aggregation_method: The aggregation method to use
    :param time_frames: The time frames that the temperature scores should be calculated for  (None to calculate all)
    :param scopes: The scopes that the temperature scores should be calculated for (None to calculate all)
    :param grouping: The names of the columns to group on
    :param scenario: The scenario to play
    :param anonymize: Whether to anonymize the resulting data set or not
    :param aggregate: Whether to aggregate the scores or not
    :return: The scores, the aggregations and the column distribution (if a
    """
    ts = TemperatureScore(
        time_frames=time_frames,
        scopes=scopes,
        default_score=default_score,
        scenario=scenario,
        grouping=grouping,
        aggregation_method=aggregation_method,
    )

    scores = ts.calculate(portfolio_data)
    aggregations = None
    if aggregate:
        aggregations = ts.aggregate_scores(scores)

    if anonymize:
        scores = ts.anonymize_data_dump(scores)

    return scores, aggregations



"""
This package helps companies and financial institutions to assess the temperature alignment of current
targets, commitments, and investment and lending portfolios, and to use this information to develop
targets for official validation by the SBTi.
"""
__version__ = "1.5.4"

from . import data
from . import utils
from . import temperature_score



from typing import Type, Optional
import pandas as pd
from ITR.configs import PortfolioCoverageTVPConfig
from ITR.portfolio_aggregation import PortfolioAggregation, PortfolioAggregationMethod


class PortfolioCoverageTVP(PortfolioAggregation):
    """
    Lookup the companies in the given portfolio and determine whether they have a SBTi approved target.

    :param config: A class defining the constants that are used throughout this class. This parameter is only required
                    if you'd like to overwrite a constant. This can be done by extending the PortfolioCoverageTVPConfig
                    class and overwriting one of the parameters.
    """

    def __init__(
        self, config: Type[PortfolioCoverageTVPConfig] = PortfolioCoverageTVPConfig
    ):
        super().__init__(config)
        self.c: Type[PortfolioCoverageTVPConfig] = config

    def get_portfolio_coverage(
        self,
        company_data: pd.DataFrame,
        portfolio_aggregation_method: PortfolioAggregationMethod,
    ) -> Optional[float]:
        """
        Get the TVP portfolio coverage (i.e. what part of the portfolio has a SBTi validated target).

        :param company_data: The company as it is returned from the data provider's get_company_data call.
        :param portfolio_aggregation_method: PortfolioAggregationMethod: The aggregation method to use
        :return: The aggregated score
        """
        company_data[self.c.OUTPUT_TARGET_STATUS] = company_data.apply(
            lambda row: 100 if row[self.c.COLS.SBTI_VALIDATED] else 0, axis=1
        )

        return self._calculate_aggregate_score(
            company_data, self.c.OUTPUT_TARGET_STATUS, portfolio_aggregation_method
        ).sum()



from enum import Enum
from datetime import date
from typing import Optional, Dict, List

import pandas as pd
import numpy as np
from pydantic import BaseModel, field_validator, Field, ValidationError


class AggregationContribution(BaseModel):
    company_name: str
    company_id: str
    temperature_score: float
    contribution_relative: Optional[float]
    contribution: Optional[float]

    def __getitem__(self, item):
        return getattr(self, item)


class Aggregation(BaseModel):
    score: float
    proportion: float
    contributions: List[AggregationContribution]

    def __getitem__(self, item):
        return getattr(self, item)


class ScoreAggregation(BaseModel):
    all: Aggregation
    influence_percentage: float
    grouped: Dict[str, Aggregation]

    def __getitem__(self, item):
        return getattr(self, item)


class ScoreAggregationScopes(BaseModel):
    S1: Optional[ScoreAggregation] = None
    S2: Optional[ScoreAggregation] = None
    S1S2: Optional[ScoreAggregation] = None
    S3: Optional[ScoreAggregation] = None
    S1S2S3: Optional[ScoreAggregation] = None

    def __getitem__(self, item):
        return getattr(self, item)


class ScoreAggregations(BaseModel):
    short: Optional[ScoreAggregationScopes] = None
    mid: Optional[ScoreAggregationScopes] = None
    long: Optional[ScoreAggregationScopes] = None

    def __getitem__(self, item):
        return getattr(self, item)


class ScenarioInterface(BaseModel):
    number: int
    engagement_type: Optional[str] 


class PortfolioCompany(BaseModel):
    company_name: str
    company_id: str
    company_isin: Optional[str] = None
    company_lei: Optional[str] = None
    investment_value: float
    engagement_target: Optional[bool] = False
    user_fields: Optional[dict] = None
    
    @field_validator("company_isin", "company_lei", mode="before")
    @classmethod
    def validate_optional_strings(cls, val):
        if pd.isnull(val) or val == "" or val == "nan":
            return None
        return val


class IDataProviderCompany(BaseModel):
    company_name: str
    company_id: str
    isic: str
    ghg_s1: Optional[float] = np.nan
    ghg_s2: Optional[float] = np.nan
    ghg_s1s2: Optional[float] = np.nan 
    ghg_s3: Optional[float] = np.nan

    # Optional fields for scope 3 categories
    ghg_s3_1: Optional[float] = np.nan
    ghg_s3_2: Optional[float] = np.nan
    ghg_s3_3: Optional[float] = np.nan
    ghg_s3_4: Optional[float] = np.nan
    ghg_s3_5: Optional[float] = np.nan
    ghg_s3_6: Optional[float] = np.nan
    ghg_s3_7: Optional[float] = np.nan
    ghg_s3_8: Optional[float] = np.nan
    ghg_s3_9: Optional[float] = np.nan
    ghg_s3_10: Optional[float] = np.nan
    ghg_s3_11: Optional[float] = np.nan
    ghg_s3_12: Optional[float] = np.nan
    ghg_s3_13: Optional[float] = np.nan
    ghg_s3_14: Optional[float] = np.nan
    ghg_s3_15: Optional[float] = np.nan

    country: Optional[str] = None
    region: Optional[str] = None
    sector: Optional[str] = None
    industry_level_1: Optional[str] = None
    industry_level_2: Optional[str] = None
    industry_level_3: Optional[str] = None
    industry_level_4: Optional[str] = None

    company_revenue: Optional[float] = 0.0
    company_market_cap: Optional[float] = 0.0
    company_enterprise_value: Optional[float] = 0.0
    company_total_assets: Optional[float] = 0.0
    company_cash_equivalents: Optional[float] = 0.0

    sbti_validated: bool = Field(
        False,
        description='True if the SBTi target status is "Target set", false otherwise',
    )
    

class SortableEnum(Enum):
    def __str__(self):
        return self.name

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) >= order.index(other)
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) > order.index(other)
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) <= order.index(other)
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            order = list(self.__class__)
            return order.index(self) < order.index(other)
        return NotImplemented


class EScope(SortableEnum):
    S1 = "S1"
    S2 = "S2"
    S1S2 = "S1+S2"
    S3 = "S3"
    S1S2S3 = "S1+S2+S3"

    @classmethod
    def get_result_scopes(cls) -> List["EScope"]:
        """
        Get a list of scopes that should be calculated if the user leaves it open.

        :return: A list of EScope objects
        """
        return [cls.S1S2, cls.S3, cls.S1S2S3]


class ETimeFrames(SortableEnum):
    SHORT = "short"
    MID = "mid"
    LONG = "long"

#test this
class ETargetReference(SortableEnum):
    ABSOLUTE = "absolute"
    INT_TO_ABS = "int_to_abs"
    INTENSITY = "intensity"
    T_SCORE = "t_score"

class S3Category(SortableEnum):
    CAT_1 = 1
    CAT_2 = 2
    CAT_3 = 3
    CAT_4 = 4
    CAT_5 = 5
    CAT_6 = 6
    CAT_7 = 7
    CAT_8 = 8
    CAT_9 = 9
    CAT_10 = 10
    CAT_11 = 11
    CAT_12 = 12
    CAT_13 = 13
    CAT_14 = 14
    CAT_15 = 15
    CAT_H_LINE = 0
    N_A = -1

class IDataProviderTarget(BaseModel):
    company_id: str
    target_type: str
    intensity_metric: Optional[str] = None
    base_year_ts: Optional[float] = None
    end_year_ts: Optional[float] = None
    scope: EScope
    #s3_category: Optional[int]
    s3_category: Optional[S3Category] = None
    coverage_s1: Optional[float] = np.nan
    coverage_s2: Optional[float] = np.nan
    coverage_s3: Optional[float] = np.nan

    reduction_ambition: Optional[float] = np.nan
    
    base_year: int
    base_year_ghg_s1: Optional[float] = np.nan
    base_year_ghg_s2: Optional[float] = np.nan
    base_year_ghg_s1s2: Optional[float] = np.nan
    base_year_ghg_s3: Optional[float] = np.nan
    
    start_year: Optional[int] = None
    end_year: int
    statement_date: Optional[date] = None
    time_frame: Optional[ETimeFrames] = None
    achieved_reduction: Optional[float] = 0.0
    to_calculate: Optional[bool] = False # Set to True if the target should be calculated

    target_ids: List[str] = Field(
        default_factory=list,
        description="""Some data providers use a unique identifier for each target. This identifier can then be used to 
        link companies targets to scores. E.g. targets MSCI1235 and MSCI4567 drive a score of 2.5° for Company 123""",
    )

    sbti_validated: bool = Field(
        False,
        description='True if the SBTi target status is "Target set", false otherwise',
    )
    
    @field_validator("intensity_metric", mode="before")
    @classmethod
    def validate_intensity_metric(cls, val):
        if pd.isnull(val) or val == "" or val == "nan":
            return None
        return val
    
    @field_validator("start_year", mode="before")
    @classmethod
    def validate_start_year(cls, val):
        if val == "" or val == "nan" or pd.isnull(val):
            return None
        return val
    
    # @field_validator("s3_category", mode="before")
    # @classmethod
    # def validate_s3_category(cls, val):
    #     if val is None:
    #         return None
    #     elif isinstance(val, S3Category):
    #         return val
    #     elif isinstance(val, int):
    #         try:
    #             return S3Category(val)
    #         except ValueError:
    #             raise ValidationError("Invalid value for s3_category")
    #     else:
    #         raise ValidationError("Invalid type for s3_category")
        
    @field_validator("target_ids", mode="before")
    @classmethod
    def convert_to_list(cls, v):
        """
        targets can be combined so target_ids field must be a list
        mode='before' is used to ensure that the validator is called before the default_factory
        with mode='before' and default_factory the users can supply single strings, e.g. "Target1"
        """
        if isinstance(v, list):
            return v
        if pd.isnull(v):
            return []
        return [v]
    
    def equals(self, other: "IDataProviderTarget") -> bool:
        """
        Check if two targets are equal.
        :param other: The other target
        :return: True if the targets are equal, False otherwise
        """
        return (self.company_id == other.company_id 
                and self.target_type == other.target_type 
                and self.scope == other.scope 
                and self.base_year == other.base_year 
                and self.end_year == other.end_year 
                and self.time_frame == other.time_frame 
                and self.target_ids == other.target_ids
        )


from abc import ABC
from enum import Enum
from typing import Type

import pandas as pd
from .configs import PortfolioAggregationConfig, ColumnsConfig
from .interfaces import EScope

class PortfolioAggregationMethod(Enum):
    """
    The portfolio aggregation method determines how the temperature scores for the individual companies are aggregated
    into a single portfolio score.
    """

    WATS = "WATS"
    TETS = "TETS"
    MOTS = "MOTS"
    EOTS = "EOTS"
    ECOTS = "ECOTS"
    AOTS = "AOTS"
    ROTS = "ROTS"

    @staticmethod
    def is_emissions_based(method: "PortfolioAggregationMethod") -> bool:
        """
        Check whether a given method is emissions-based (i.e. it uses the emissions to calculate the aggregation).

        :param method: The method to check
        :return:
        """
        return (
            method == PortfolioAggregationMethod.MOTS
            or method == PortfolioAggregationMethod.EOTS
            or method == PortfolioAggregationMethod.ECOTS
            or method == PortfolioAggregationMethod.AOTS
            or method == PortfolioAggregationMethod.ROTS
        )

    @staticmethod
    def get_value_column(
        method: "PortfolioAggregationMethod", column_config: Type[ColumnsConfig]
    ) -> str:
        map_value_column = {
            PortfolioAggregationMethod.MOTS: column_config.MARKET_CAP,
            PortfolioAggregationMethod.EOTS: column_config.COMPANY_ENTERPRISE_VALUE,
            PortfolioAggregationMethod.ECOTS: column_config.COMPANY_EV_PLUS_CASH,
            PortfolioAggregationMethod.AOTS: column_config.COMPANY_TOTAL_ASSETS,
            PortfolioAggregationMethod.ROTS: column_config.COMPANY_REVENUE,
        }

        return map_value_column.get(method, column_config.MARKET_CAP)


class PortfolioAggregation(ABC):
    """
    This class is a base class that provides portfolio aggregation calculation.

    :param config: A class defining the constants that are used throughout this class. This parameter is only required
                    if you'd like to overwrite a constant. This can be done by extending the PortfolioAggregationConfig
                    class and overwriting one of the parameters.
    """

    def __init__(
        self, config: Type[PortfolioAggregationConfig] = PortfolioAggregationConfig
    ):
        self.c = config

    def _check_column(self, data: pd.DataFrame, column: str):
        """
        Check if a certain column is filled for all companies. If not throw an error.
        :param data: The data to check
        :param column: The column to check
        :return:
        """
        missing_data = data[pd.isnull(data[column])][self.c.COLS.COMPANY_NAME].unique()
        if len(missing_data):
            if (
                column == self.c.COLS.GHG_SCOPE1  
                or column == self.c.COLS.GHG_SCOPE2  
                or column == self.c.COLS.GHG_SCOPE3
            ):
                raise ValueError(
                    f"A value for {column} is needed for all aggregation methods except for WATS. \n"
                    f"So please try to estimate appropriate values or remove these companies from the aggregation calculation: {','.join(missing_data)}"            
                )
            else:
                raise ValueError(
                    f"The value for {column} is missing for the following companies: {','.join(missing_data)}"                      
                )
                

    def _calculate_aggregate_score(
        self,
        data: pd.DataFrame,
        input_column: str,
        portfolio_aggregation_method: PortfolioAggregationMethod,
    ) -> pd.Series:
        """
        Aggregate the scores in a given column based on a certain portfolio aggregation method.

        :param data: The data to run the calculations on
        :param input_column: The input column (containing the scores)
        :param portfolio_aggregation_method: The method to use
        :return: The aggregates score
        """
        if portfolio_aggregation_method == PortfolioAggregationMethod.WATS:
            total_investment_weight = data[self.c.COLS.INVESTMENT_VALUE].sum()
            try:
                return data.apply(
                    lambda row: (row[self.c.COLS.INVESTMENT_VALUE] * row[input_column])
                    / total_investment_weight,
                    axis=1,
                )
            except ZeroDivisionError:
                raise ValueError("The portfolio weight is not allowed to be zero")

        # Total emissions weighted temperature score (TETS)
        elif portfolio_aggregation_method == PortfolioAggregationMethod.TETS:
            use_S1S2 = data[self.c.COLS.SCOPE].isin([EScope.S1S2, EScope.S1S2S3])
            use_S3 = data[self.c.COLS.SCOPE].isin([EScope.S3, EScope.S1S2S3])
            if use_S3.any():
                self._check_column(data, self.c.COLS.GHG_SCOPE3)
            if use_S1S2.any():
                self._check_column(data, self.c.COLS.GHG_SCOPE12)
            # Calculate the total emissions of all companies
            emissions = (use_S1S2 * data[self.c.COLS.GHG_SCOPE12]).sum() + (
                use_S3 * data[self.c.COLS.GHG_SCOPE3]
            ).sum()
            try:
                weighted_scores = (
                    (use_S1S2 * data[self.c.COLS.GHG_SCOPE12] + use_S3 * data[self.c.COLS.GHG_SCOPE3])
                    / emissions
                    * data[input_column]
                )
                return weighted_scores
            except ZeroDivisionError:
                raise ValueError("The total emissions should be higher than zero")

        elif PortfolioAggregationMethod.is_emissions_based(
            portfolio_aggregation_method
        ):
            # These four methods only differ in the way the company is valued.
            if portfolio_aggregation_method == PortfolioAggregationMethod.ECOTS:
                self._check_column(data, self.c.COLS.COMPANY_ENTERPRISE_VALUE)
                self._check_column(data, self.c.COLS.CASH_EQUIVALENTS)
                data[self.c.COLS.COMPANY_EV_PLUS_CASH] = (
                    data[self.c.COLS.COMPANY_ENTERPRISE_VALUE]
                    + data[self.c.COLS.CASH_EQUIVALENTS]
                )

            value_column = PortfolioAggregationMethod.get_value_column(
                portfolio_aggregation_method, self.c.COLS
            )

            # Calculate the total owned emissions of all companies
            try:
                self._check_column(data, self.c.COLS.INVESTMENT_VALUE)
                self._check_column(data, value_column)
                use_S1S2 = data[self.c.COLS.SCOPE].isin([EScope.S1S2, EScope.S1S2S3])
                use_S3 = data[self.c.COLS.SCOPE].isin([EScope.S3, EScope.S1S2S3])
                if use_S1S2.any():
                    self._check_column(data, self.c.COLS.GHG_SCOPE12)
                if use_S3.any():
                    self._check_column(data, self.c.COLS.GHG_SCOPE3)
                data[self.c.COLS.OWNED_EMISSIONS] = (
                    data[self.c.COLS.INVESTMENT_VALUE] / data[value_column]
                ) * (
                    use_S1S2 * data[self.c.COLS.GHG_SCOPE12]
                    + use_S3 * data[self.c.COLS.GHG_SCOPE3]
                )
            except ZeroDivisionError:
                raise ValueError(
                    "To calculate the aggregation, the {} column may not be zero".format(
                        value_column
                    )
                )
            owned_emissions = data[self.c.COLS.OWNED_EMISSIONS].sum()
            
            try:
                # Calculate the MOTS value per company
                weighted_scores = data.apply(
                    lambda row: (row[self.c.COLS.OWNED_EMISSIONS] / owned_emissions)
                    * row[input_column],
                    axis=1,
                )
                return weighted_scores
            
            except ZeroDivisionError:
                raise ValueError("The total owned emissions can not be zero")
        else:
            raise ValueError("The specified portfolio aggregation method is invalid")



from typing import Type, List
from pydantic import ValidationError
from datetime import date
import logging

import pandas as pd
import numpy as np
from ITR.data.data_provider import DataProvider
from ITR.configs import ColumnsConfig
from ITR.interfaces import IDataProviderCompany, IDataProviderTarget, S3Category


class ExcelProvider(DataProvider):
    """
    Data provider skeleton for CSV files. This class serves primarily for testing purposes only!

    :param config: A dictionary containing a "path" field that leads to the path of the CSV file
    """

    def __init__(self, path: str, config: Type[ColumnsConfig] = ColumnsConfig):
        super().__init__()
        self.data = pd.read_excel(path, sheet_name=None, skiprows=0)
        # Set missing values to 0.0
        self.data['target_data'][['achieved_reduction']] = self.data['target_data'][['achieved_reduction']].fillna(0.0)
        self.data['target_data']['end_year'] = self.data['target_data']['end_year'].fillna(0)
        self.data['target_data']['scope'] = self.data['target_data']['scope'].replace({'S1S2S3': 'S1+S2+S3'})
        self.data['target_data']['scope'] = self.data['target_data']['scope'].replace({'S1S2': 'S1+S2'})
        # Check for optional columns in 'fundamental_data' and handle missing values
        optional_columns = ['ghg_s1', 'ghg_s2', 'ghg_s1s2', 'ghg_s3']
        for col in optional_columns:
            if col not in self.data['fundamental_data'].columns:
                self.data['fundamental_data'][col] = 0.0
            else:
                self.data['fundamental_data'][col] = self.data['fundamental_data'][col].fillna(0.0)
       
        try:
           self.data['target_data']['s3_category'] = self.data['target_data'].apply(self._process_row, axis=1)
        except Exception as e:
            print(f"An error occurred: {e}")
       
        self.c = config

    def _process_row(self, row):
        if row['scope'] in ['S3', 'S1+S2+S3']:
            if pd.isna(row['s3_category']) or row['s3_category'] == 0:
                return 0
            elif 1 <= int(row['s3_category']) <= 15:
                return int(row['s3_category'])
            else:
                return 0
        else:
            return -1

    def get_targets(self, company_ids: List[str]) -> List[IDataProviderTarget]:
        """
        Get all relevant targets for a list of company ids (ISIN). This method should return a list of
        IDataProviderTarget instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the targets
        """
        model_targets = self._target_df_to_model(self.data["target_data"])
        model_targets = [
            target for target in model_targets if target.company_id in company_ids
        ]
        return model_targets

    def _target_df_to_model(self, df_targets):
        """
        transforms target Dataframe into list of IDataProviderTarget instances

        :param df_targets: pandas Dataframe with targets
        :return: A list containing the targets
        """
        logger = logging.getLogger(__name__)
        df_targets = df_targets.copy()  # Avoid mutating the original DataFrame
         # 1) Check if 'statement_date' looks like a date
        df_targets['statement_date'] = pd.to_datetime(df_targets['statement_date'], format='%Y', errors='coerce')

        # 2) If 'statement_date' is empty, check 'start_year'
        df_targets.loc[df_targets['statement_date'].isna(), 'statement_date'] = pd.to_datetime(df_targets['start_year'], format='%Y', errors='coerce')

        # 3) If 'start_year' is empty, use 'base_year'
        df_targets.loc[df_targets['statement_date'].isna(), 'statement_date'] = pd.to_datetime(df_targets['base_year'], format='%Y', errors='coerce')
        targets = df_targets.to_dict(orient="records")
        model_targets: List[IDataProviderTarget] = []
        for target in targets:
            try:
                # Replace NaN/NaT with None so Pydantic handles Optional fields correctly
                cleaned = {}
                for k, v in target.items():
                    try:
                        if v is None or (not isinstance(v, (list, dict, str)) and pd.isna(v)):
                            cleaned[k] = None
                        else:
                            cleaned[k] = v
                    except (ValueError, TypeError):
                        cleaned[k] = v
                target = cleaned
                # Convert float years to int (Excel reads int columns as float)
                for year_col in ['base_year', 'end_year', 'start_year']:
                    if year_col in target and target[year_col] is not None:
                        target[year_col] = int(target[year_col])
                # Map the s3_category values to the S3Category enum
                if 's3_category' in target and target['s3_category'] is not None:
                    target['s3_category'] = S3Category(int(target['s3_category']))
                model_targets.append(IDataProviderTarget.model_validate(target))
            except (ValidationError, TypeError) as e:
                logger.warning(
                    f"Target validation failed and will be skipped: {e}"
                )
        return model_targets

    def get_company_data(self, company_ids: List[str]) -> List[IDataProviderCompany]:
        """
        Get all relevant data for a list of company ids (ISIN). This method should return a list of IDataProviderCompany
        instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the company data
        """
        data_company = self.data["fundamental_data"]
        # Pydantic doesn't accept NaN values in string fields, 
        # so we need to convert them to empty strings
        fields_to_convert = ['isic','country', 'sector', 'industry_level_1', 'industry_level_2', 
                             'industry_level_3', 'industry_level_4']
        for field in fields_to_convert:
          if field in data_company.columns:
            data_company[field] = data_company[field].fillna('')

        companies = data_company.to_dict(orient="records")
        model_companies: List[IDataProviderCompany] = [
            IDataProviderCompany.model_validate(company) for company in companies
        ]
        for company in model_companies:
            if company.ghg_s1 is not None or company.ghg_s2 is not None:
                company.ghg_s1s2 = company.ghg_s1 + company.ghg_s2  # type: ignore (data set to 0.0 when reading)

        model_companies = [
            target for target in model_companies if target.company_id in company_ids
        ]
        return model_companies

    def get_sbti_targets(self, companies: list) -> list:
        """
        For each of the companies, get the status of their target (Target set, Committed or No target) as it's known to
        the SBTi.

        :param companies: A list of companies. Each company should be a dict with a "company_name" and "company_id"
                            field.
        :return: The original list, enriched with a field called "sbti_target_status"
        """
        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import List

from ITR.interfaces import IDataProviderCompany, IDataProviderTarget


class DataProvider(ABC):
    """
    General data provider super class.
    """

    def __init__(self, **kwargs):
        """
        Create a new data provider instance.

        :param config: A dictionary containing the configuration parameters for this data provider.
        """
        pass

    @abstractmethod
    def get_targets(self, company_ids: List[str]) -> List[IDataProviderTarget]:
        """
        Get all relevant targets for a list of company ids (ISIN). This method should return a list of
        IDataProviderTarget instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the targets
        """
        raise NotImplementedError

    @abstractmethod
    def get_company_data(self, company_ids: List[str]) -> List[IDataProviderCompany]:
        """
        Get all relevant data for a list of company ids (ISIN). This method should return a list of IDataProviderCompany
        instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the company data
        """
        raise NotImplementedError

    @abstractmethod
    def get_sbti_targets(self, companies: list) -> list:
        """
        For each of the companies, get the status of their target (Target set, Committed or No target) as it's known to
        the SBTi.

        :param companies: A list of companies. Each company should be a dict with a "company_name" and "company_id"
                            field.
        :return: The original list, enriched with a field called "sbti_target_status"
        """
        raise NotImplementedError


class CompanyNotFoundException(Exception):
    """
    This exception occurs when a company is not found.
    """
    pass



# This is target_validation.py
import datetime
import itertools
import logging
import copy

import pandas as pd
from typing import Type, List, Tuple, Optional, cast
from ITR.configs import PortfolioAggregationConfig, TemperatureScoreConfig

from ITR.interfaces import (
    IDataProviderTarget,
    IDataProviderCompany,
    EScope,
    ETimeFrames,
    S3Category
)

logger = logging.getLogger(__name__)


class TargetProtocol:
    """
    This class validates the targets, to make sure that only active, useful 
    targets are considered. It then combines the targets with company-related 
    data into a dataframe where there's one row for each of the nine possible 
    target types (short, mid, long * S1+S2, S3, S1+S2+S3). 
    This class follows the procedures outlined by the target protocol that is 
    a part of the "Temperature Rating Methodology" (2024), which has been created 
    by CDP Worldwide and WWF International.

    :param config: A Portfolio aggregation config
    """

    def __init__(
        self, config: Type[PortfolioAggregationConfig] = PortfolioAggregationConfig
    ):
        self.c = config
        self.logger = logging.getLogger(__name__)
        self.target_data: pd.DataFrame = pd.DataFrame()
        self.intensity_metric_types = TemperatureScoreConfig.intensity_metric_types
        self.company_data: pd.DataFrame = pd.DataFrame()
        self.data: pd.DataFrame = pd.DataFrame()

    def process(
        self, targets: List[IDataProviderTarget], companies: List[IDataProviderCompany]
    ) -> pd.DataFrame:
        """
        Process the targets and companies, validate all targets and return a data frame that combines all targets and company data into a 9-box grid.

        :param targets: A list of targets
        :param companies: A list of companies
        :return: A data frame that combines the processed data
        """
        logger.info(f"started processing {len(targets)=} and {len(companies)=}")
        # Create multiindex on company, timeframe and scope for performance later on
        self.company_data = pd.DataFrame.from_records([c.model_dump() for c in companies])
        targets = self._prepare_targets(targets)
        self.target_data = pd.DataFrame.from_records([c.model_dump() for c in targets])
        if self.target_data.empty:
            logger.warning("No valid targets after preparation")
            # Return an empty merge so downstream code gets the right columns
            self.data = pd.DataFrame(columns=[self.c.COLS.COMPANY_ID, self.c.COLS.TIME_FRAME, self.c.COLS.SCOPE])
            return pd.merge(
                left=self.data, right=self.company_data, how="outer", on=["company_id"]
            )
        self.target_data['statement_date'] = pd.to_datetime(self.target_data['statement_date'])

        # Create an indexed DF for performance purposes
        self.target_data.index = (
            self.target_data.reset_index()
            .set_index(
                [self.c.COLS.COMPANY_ID, self.c.COLS.TIME_FRAME, self.c.COLS.SCOPE]
            )
            .index
        )
        self.target_data = self.target_data.sort_index()

        self.target_data = self.sort_on_vintage(self.target_data)

        self.target_data = self.sort_boundary_coverage(self.target_data)

        self.target_data['reduction_ambition'] = self.target_data.apply(self._scale_reduction_ambition_by_boundary_coverage_new, axis=1)

        self.group_targets()
        
        #Drop the column sbti_validated from target_data since it's no longer needed
        self.data = self.data.drop(columns=["sbti_validated"], errors="ignore")
        out =  pd.merge(
            left=self.data, right=self.company_data, how="outer", on=["company_id"]
        )
        logger.info(f"finished processing {out.shape=}")
        return out

    def _validate(self, target: IDataProviderTarget) -> bool:
        """
        Validate a target, meaning it should:

        * Have a valid type
        * Not be finished
        * A valid end year

        :param target: The target to validate
        :return: True if it's a valid target, false if it isn't
        """
        # If the target is set using the CDP_WWF temperature scoring methodology
        if target.target_type.lower() == "t_score":
            return self._validate_t_score(target)
        # Only absolute targets or intensity targets with a valid intensity metric are allowed.
        # As of v1.5 we have target type 'int_to_abs' which is an intensity target that has been converted to an absolute target
        target_type = (
            "abs" in target.target_type.lower()
            or "int_to_abs" in target.target_type.lower()
            or (
                "intensity" in target.target_type.lower()
                and target.intensity_metric in self.intensity_metric_types
            )
        )
        # There must be a number in the reduction ambition field, even zero is okay.
        target_reduction_ambition = not pd.isna(target.reduction_ambition)

        # The target should not have achieved its reduction yet.
        target_progress = (
            pd.isnull(target.achieved_reduction)
            or target.achieved_reduction is None
            or target.achieved_reduction < 1
        )

        # The end year should be greater than the start year.
        if target.start_year is None or pd.isnull(target.start_year):
            target.start_year = target.base_year

        target_end_year = target.end_year > target.start_year

        # The end year should be greater than or equal to the current year
        target_current = target.end_year >= datetime.datetime.now().year

        # Check that base year ghg data is available for the scope of the target
        s1 = target.scope != EScope.S1 or (
            pd.notna(target.coverage_s1)
            and pd.notna(target.base_year_ghg_s1) 
        )
        s2 = target.scope != EScope.S2 or (
            pd.notna(target.coverage_s2)
            and pd.notna(target.base_year_ghg_s2)  
        )
        s1s2 = target.scope != EScope.S1S2 or (
            pd.notna(target.base_year_ghg_s1)
            and pd.notna(target.base_year_ghg_s2) 
        )
        # Note that base year s3 ghg is checked in the method _find_s3_targets
        # for individual s3 category targets that have been split from s1s2s3 targets
        s1s2s3 = target.scope != EScope.S1S2S3 or (
            pd.notna(target.base_year_ghg_s1)
            and pd.notna(target.base_year_ghg_s2) 
        )
        s3 = target.scope != EScope.S3 or (
            pd.notna(target.coverage_s3)
            and pd.notna(target.base_year_ghg_s3)
        )
        return (
            target_type
            and target_reduction_ambition 
            and target_progress 
            and target_end_year 
            and target_current 
            and s1 
            and s2
            and s1s2
            and s1s2s3
            and s3
        )
    
    def _validate_t_score(self, target: IDataProviderTarget) -> bool:
        """
        Validate targets set using the CDP_WWF temperature scoing methodology
        :param target: The target to validate
        :return: True if it's a valid target, false if it isn't
        """
        # The end year should be greater than the start year.
        if target.start_year is None or pd.isnull(target.start_year):
            target.start_year = target.base_year

        target_end_year = target.end_year > target.start_year

        # The end year should be greater than or equal to the current year
        # Added in update Oct 22
        target_current = target.end_year >= datetime.datetime.now().year

        # Target scope must be S3
        target_scope = target.scope == EScope.S3
        
        return (
            target_end_year
            and target_current
            and target_scope
        )

    def _split_s1s2s3(self,
        target: IDataProviderTarget,
    ) -> Tuple[IDataProviderTarget, Optional[IDataProviderTarget]]:
        """
        If there is a s1s2s3 scope, split it into two targets with s1s2 and s3
        This S3 target becomes a headline target
        :param target: The input target
        :return The split targets or the original target and None
        """
        if target.target_type.lower() == "t_score":
            return target, None
        if target.scope == EScope.S1S2S3:
            s1s2, s3 = target.copy(), None
            s1s2.s3_category = S3Category.N_A
            # Assign scalar values to the new DataFrames
            s1s2.base_year_ghg_s1 = cast(float, target.base_year_ghg_s1)
            s1s2.base_year_ghg_s2 = cast(float, target.base_year_ghg_s2)
            if (
                not pd.isnull(target.base_year_ghg_s1)
                and not pd.isnull(target.base_year_ghg_s2)
            ) or target.coverage_s1 == target.coverage_s2:
                s1s2.scope = EScope.S1S2
                if (
                    pd.notna(s1s2.coverage_s1)
                    and pd.notna(s1s2.coverage_s2)
                    and pd.notna(s1s2.base_year_ghg_s1)
                    and pd.notna(s1s2.base_year_ghg_s2)
                    and s1s2.base_year_ghg_s1 + s1s2.base_year_ghg_s2 != 0
                ):
                    coverage_percentage = (
                        s1s2.coverage_s1 * s1s2.base_year_ghg_s1
                        + s1s2.coverage_s2 * s1s2.base_year_ghg_s2
                    ) / (s1s2.base_year_ghg_s1 + s1s2.base_year_ghg_s2)
                    s1s2.coverage_s1 = coverage_percentage
                    s1s2.coverage_s2 = coverage_percentage

            if not pd.isnull(target.coverage_s3):
                s3 = target.copy()
                s3.scope = EScope.S3
                if s3.s3_category == S3Category.N_A:
                    s3.s3_category = S3Category.CAT_H_LINE
            return s1s2, s3
                
        else:
            return target, None
           
    def _split_s1s2_new(self,
        target: pd.DataFrame,
    ) -> pd.DataFrame:
        """
        Split the target into two targets, one for the S1 data and one for the S2 data.

        :param target: The target to potentially split.
        :return: A list containing the S1 target and the S2 target from the split.
        """
        if target[self.c.COLS.TARGET_REFERENCE_NUMBER].iloc[0].lower() == "t_score":
            return target
        targets = target.iloc[0].copy()
        # before splitting S1S2 targets we need to verify that there is GHG data to aggregate the scores later
        # TODO - verify that company is one unique row - for future cleaning module
        company = self.company_data[self.company_data[self.c.COLS.COMPANY_ID] == targets.company_id]
        if (not (pd.isna(company[self.c.COLS.GHG_SCOPE1].values[0])
            or pd.isna(company[self.c.COLS.GHG_SCOPE2].values[0]))
            and target.scope.iloc[0] == EScope.S1S2
        ):
            s1 = targets.copy()
            s2 = targets.copy()           

            # Assign scalar values to the new DataFrames
            s1['coverage_s1'] = targets['coverage_s1']
            s2['coverage_s2'] = targets['coverage_s2']
            s1['reduction_ambition'] = targets['reduction_ambition']
            s2['reduction_ambition'] = targets['reduction_ambition']
        
            s1['scope'] = EScope.S1
            s2['scope'] = EScope.S2
            # Append '_1' and '_2' to the target_ids of s1 and s2, respectively
            s1['target_ids'] = [id_ + '_1' for id_ in s1['target_ids']]
            s2['target_ids'] = [id_ + '_2' for id_ in s2['target_ids']]
      
            return pd.DataFrame([s1, s2])
        return target
        
    @staticmethod
    def _scale_reduction_ambition_by_boundary_coverage_new(
        target: pd.DataFrame,
        ) -> float:
        """
        Change in ITR method 1.5: all targets have their ambition scaled by their boundary coverage.
        :param target: The input target
        :return: The original target with a weighted reduction ambition, if so required
        """
                
        if pd.isna(target.reduction_ambition):
            reduction_ambition = 0.0
        
        elif target.scope == EScope.S1:
            reduction_ambition = (
                target.reduction_ambition * target.coverage_s1
            )
        elif target.scope == EScope.S2:
            reduction_ambition = (
                target.reduction_ambition * target.coverage_s2
            )
        elif target.scope == EScope.S3:
            reduction_ambition = (
                target.reduction_ambition * target.coverage_s3
            )
        elif target.scope == EScope.S1S2:
            if pd.isna(target.coverage_s1):
                target.coverage_s1 = 0.0
            if pd.isna(target.coverage_s2):
                target.coverage_s2 = 0.0
            if (
                not pd.isnull(target.base_year_ghg_s1)
                and not pd.isnull(target.base_year_ghg_s2)
                and target.base_year_ghg_s1 + target.base_year_ghg_s2 != 0):
                    combined_coverage = (
                        target.coverage_s1 * target.base_year_ghg_s1
                        + target.coverage_s2 * target.base_year_ghg_s2
                    ) / (target.base_year_ghg_s1 + target.base_year_ghg_s2)
  
                    target.coverage_s1 = combined_coverage
                    target.coverage_s2 = combined_coverage
                    reduction_ambition = (
                        target.reduction_ambition * combined_coverage
                    )
            else:
                reduction_ambition = 0.0
        else:
            reduction_ambition = target.reduction_ambition

        return reduction_ambition # type: ignore
    @staticmethod
    def _assign_time_frame(target: IDataProviderTarget) -> IDataProviderTarget:
        """
        Time frame is forward looking: target year - current year. 
            Less than 5y = short, 
            between 5 and 10 is mid, 
            more than 10 is long

        :param target: The input target
        :return: The original target with the time_frame field filled out (if so required)
        """
        now = datetime.datetime.now()
        time_frame = target.end_year - now.year
        # Method 6.3.6.1: if target is of type T_SCORE and 
        # the company is validated by sbti, then add five years to timeframe
        if (target.target_type.lower() == "t_score") and target.sbti_validated:
            time_frame += 5

        if time_frame < 5:
            target.time_frame = ETimeFrames.SHORT
        elif time_frame <= 10:
            target.time_frame = ETimeFrames.MID
        elif time_frame > 10:
            target.time_frame = ETimeFrames.LONG

        return target

    def _prepare_targets(self, targets: List[IDataProviderTarget]):
        """
        logic
            - drop invalid targets
            - identifying the pure-S2 targets for later use
            - splitting s1s2s3 into s1s2 and s3
            - combining s1 and s2
            - assign target.reduction_ambition by considering target's boundary coverage
 
        :param targets:
        :return:
        """
        target_input_count = len(targets)
        targets = list(filter(self._validate, targets))
        logger.info(f"dropped {(target_input_count - len(targets))=:,} invalid targets")
        targets = [self._assign_time_frame(target) for target in targets]

        targets = list(
            filter(
                None, itertools.chain.from_iterable(map(self._split_s1s2s3, targets))
            )
        )
        
        return targets

    def _find_target(self, row: pd.Series, target_columns: List[str]) -> pd.DataFrame:
       
        """
        Find the target that corresponds to a given row. If there are multiple targets available, filter them.

        :param row: The row from the data set that should be looked for
        :param target_columns: The columns to return
        :return: records from the input data, which contains company and target information, that meet specific criteria. For example, record of greatest emissions_in_scope
        """
        try:

            target_data = self.target_data.xs(
                (
                    row[self.c.COLS.COMPANY_ID],
                    row[self.c.COLS.TIME_FRAME],
                    row[self.c.COLS.SCOPE],
                )
            ).copy() # type: ignore
            if isinstance(target_data, pd.Series):
                # One match with Target data
                result_df = pd.DataFrame([target_data], columns=target_columns)

            else:
                if target_data.scope.iloc[0] == EScope.S3:
                    coverage_column = self.c.COLS.COVERAGE_S3
                elif target_data.scope.iloc[0] == EScope.S2:
                    coverage_column = self.c.COLS.COVERAGE_S2
                else:
                    coverage_column = self.c.COLS.COVERAGE_S1
                # In case more than one target is available; we prefer targets with 
                # later confirmation date, 
                # higher coverage,
                # later end year, and target type 'absolute'
                
                # We also prefer longer time spans within the same time frame
                target_data['END_YEAR_MINUS_BASE_YEAR'] = (
                    target_data[self.c.COLS.END_YEAR] 
                    - target_data[self.c.COLS.BASE_YEAR]
                )
                # Reduction ambition is measured by CAR
                try:
                    target_data['CAR'] = target_data.apply(
                        lambda row: (
                            1.0 if row[self.c.COLS.REDUCTION_AMBITION] >= 1 else abs(
                                (1 - row[self.c.COLS.REDUCTION_AMBITION]) ** 
                                (1 / row['END_YEAR_MINUS_BASE_YEAR']) - 1
                            )
                        ),
                        axis=1
                    )
                except ZeroDivisionError:
                    target_data['CAR'] = 0.0
                # Scope 3 targets need to be filtered separately
                if target_data.scope.iloc[0] != EScope.S3:
                    target_data = (
                        target_data.sort_values(
                            by=[
                                self.c.COLS.TARGET_CONFIRM_DATE,
                                coverage_column,
                                self.c.COLS.TARGET_REFERENCE_NUMBER,
                                'CAR',
                                'END_YEAR_MINUS_BASE_YEAR',
                                self.c.COLS.END_YEAR,
                            ],
                            axis=0,
                            ascending=[False, False, True, False,False, False ],
                        ).iloc[0][target_columns]
                    )
                    result_df = pd.DataFrame([target_data], columns=target_columns)

                else:

                    result_df = self._find_s3_targets(target_data, target_columns)
                    if result_df.empty:
                        result_df = pd.DataFrame([row], columns=target_columns)
                        result_df['to_calculate'] = False 
                        return result_df

            result_df.loc[:, 'to_calculate'] = True
            return result_df
                          
        except KeyError:
            # No target found
            result_df = pd.DataFrame([row], columns=target_columns)
            result_df['to_calculate'] = False # TS for selected targets are not to be calculated
            return result_df
    
    def _find_s3_targets(self, target_data: pd.DataFrame, target_columns: List[str]) -> pd.DataFrame:
        """
        Find S3 target that correspond to the given row. Note that there may be more
        than one S3 target. We first look for a headline target and return that. 
        If there is none we look for the targets with the latest confirmation date.
        Then check if there is more than one target with the same confirmation date.
        The method then returns all non headline targets with the latest confirmation date.

        :param target_data: The target data
        :param target_columns: The columns to return
        :return: The target data that meet the criteria
        """
        selected_targets = []
        headline_target = target_data[target_data['s3_category'] == S3Category.CAT_H_LINE]
        if not headline_target.empty:
            # Sort by vintage and cope 3 coverage 
            headline_target = headline_target.sort_values(
                                        by=[
                                            self.c.COLS.TARGET_CONFIRM_DATE, 
                                            self.c.COLS.COVERAGE_S3,
                                            self.c.COLS.TARGET_REFERENCE_NUMBER,
                                            'CAR',
                                            'END_YEAR_MINUS_BASE_YEAR',
                                            ], 
                                            axis=0,
                                            ascending=[False, False, True, False, False])           
        
            selected_targets.append(headline_target.head(1)) 
        else: # If there is no headline target, we loop through the categories and select the best target
            filtered_targets = target_data[target_data['s3_category'].isin(
                [cat for cat in S3Category if cat not in [S3Category.CAT_H_LINE, S3Category.N_A]]
                )]
            # Invalidate targets with no base year GHG data
            filtered_targets = filtered_targets.dropna(subset=['base_year_ghg_s3'])
            if filtered_targets.empty:
                return pd.DataFrame(columns=target_columns)
            sorted_targets = filtered_targets.sort_values(
                                    by=[
                                        self.c.COLS.TARGET_CONFIRM_DATE, 
                                        self.c.COLS.COVERAGE_S3, 
                                        self.c.COLS.TARGET_REFERENCE_NUMBER,
                                        'CAR',
                                        'END_YEAR_MINUS_BASE_YEAR',
                                        ],
                                        ascending=[False, False, True, False, False])
           
        # Select the top target for each category
            top_targets = sorted_targets.groupby('s3_category').head(1)
            selected_targets.append(top_targets)

        if selected_targets:
            selected_targets_df = pd.concat(selected_targets, ignore_index=True)
            final = selected_targets_df[target_columns]
        else:
            final = pd.DataFrame(columns=target_columns)
    
        return final
    
    def group_targets(self):
        """
        Group the targets and create the target grid (short, mid, long * s1s2, s3, s1s2s3).
        Group valid targets by category & filter multiple targets
        Input: a list of valid targets for each company:
        For each company:

        Group all valid targets based on scope (S1+S2 / S3 / S1+S2+S3) and time frame (short / mid / long-term)
        into 6 categories.

        For each category: if more than 1 target is available, filter based on the following criteria
        -- Latest vintage
        -- Highest boundary coverage
        -- Latest end year
        -- Target type: Absolute over intensity
        -- If all else is equal: average the ambition of targets
        """
        self.target_data = self.target_data.sort_index(level=self.target_data.index.names)

        grid_columns = [
            self.c.COLS.COMPANY_ID,
            self.c.COLS.TIME_FRAME,
            self.c.COLS.SCOPE,
        ]
        companies = self.company_data[self.c.COLS.COMPANY_ID].unique()
        #scopes = [EScope.S1S2, EScope.S3, EScope.S1S2S3]
        scopes = [EScope.S1, EScope.S2, EScope.S1S2, EScope.S3, EScope.S1S2S3]
        empty_columns = [
            column for column in self.target_data.columns if column not in grid_columns
        ]
        extended_data = pd.DataFrame(
            list(
                itertools.product(
                    *[companies, ETimeFrames, scopes] + [[None]] * len(empty_columns)
                )
            ),
            columns=grid_columns + empty_columns,
        )

        target_columns = extended_data.columns.tolist() 
        results = []
        for _, row in extended_data.iterrows():
            result = self._find_target(row, target_columns)
            results.append(result)
        # Remove empty or all-NA columns from each DataFrame in results (pandas futureproofing)
        results = [df.dropna(how='all', axis=1) for df in results]
        self.data = pd.concat(results, ignore_index=True)

        # Restore any schema columns that were dropped by dropna
        for col in target_columns:
            if col not in self.data.columns:
                self.data[col] = None

        return self.data
    
    def sort_on_vintage(self, target_data: pd.DataFrame) -> pd.DataFrame:
        """
        For each combination of company id, time frame, and scope, find all targets with the same latest 
        TARGET_CONFIRM_DATE and return as a DataFrame of targets.

        :param target_data: a DataFrame with target data
        :return: A DataFrame of targets sorted based on their vintage
        """
        target_data['statement_year'] = target_data['statement_date'].dt.year

        # Find the latest year for each group
        latest_years = target_data.groupby(
            level=[self.c.COLS.COMPANY_ID, self.c.COLS.TIME_FRAME, self.c.COLS.SCOPE]
        )['statement_year'].transform('max')

        # Keep rows matching the latest year, or all rows in groups where all dates are NaN
        latest_data = target_data[
            (target_data['statement_year'] == latest_years) | latest_years.isna()
        ]

        latest_data = latest_data.drop(columns=['statement_year'])

        return latest_data
    
    def sort_boundary_coverage(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        We want to select the scope 1 and scope 2 targets that have the highest
        combined boundary coverage for each group of self.c.COLS.COMPANY_ID, 
        self.c.COLS.TIME_FRAME, self.c.COLS.SCOPE. 
        We compare combinations of individual scope 1 and scope 2
        targets with combined S1+S2 targets, and select the combination with 
        the highest boundary coverage. If there is only one single scope target
        (e.g., S1 but no S2), we dismiss it.

        :param data: A dataframe with the target data
        :return: A dataframe with the target data sorted by boundary coverage
        """
        # Save the original columns to use when returning the result
        original_columns = data.columns.tolist()

        # Need to save dypes as well since conversion from Series to DataFrame can change them
        original_dtype = data.dtypes
        scope_1_2_mask = data[self.c.COLS.SCOPE].isin([EScope.S1, EScope.S2, EScope.S1S2])

        # Use the boolean mask to filter for Scope 1 and Scope 2
        scope_1_2_data = data[scope_1_2_mask]

        # If there are no S1/S2/S1S2 targets, nothing to sort
        if scope_1_2_data.empty:
            return data

        # Group by COMPANY_ID and TIME_FRAME
        grouped_data = scope_1_2_data.groupby(level=[self.c.COLS.COMPANY_ID, self.c.COLS.TIME_FRAME])
                
        def get_best_s1_s2_combination(group: pd.DataFrame) -> Tuple[float, pd.DataFrame]:
            """
            Find the combination of S1 and S2 targets with the highest combined coverage
            :param group: The group of data to analyze
            :return: The row with the highest combined coverage
            """
            group_reset = group.reset_index(drop=True)
            ghg_s1 = self.c.COLS.BASEYEAR_GHG_S1
            ghg_s2 = self.c.COLS.BASEYEAR_GHG_S2

            # Filter the data for S1 and S2 scopes
            idx_s1 = group_reset[group_reset[self.c.COLS.SCOPE] == EScope.S1]
            idx_s2 = group_reset[group_reset[self.c.COLS.SCOPE] == EScope.S2]

            max_coverage = -float('inf')
            best_combination = None

            # Iterate over all combinations of S1 and S2
            for i, row_s1 in idx_s1.iterrows():
                for j, row_s2 in idx_s2.iterrows():
                    # Calculate the weighted coverage
                    weighted_coverage = (
                        (row_s1['coverage_s1'] * row_s1[ghg_s1] + row_s2['coverage_s2'] * row_s2[ghg_s2]) 
                        / (row_s1[ghg_s1] + row_s2[ghg_s2])
                    )

                    # Check if this is the best combination so far
                    if weighted_coverage > max_coverage:
                        max_coverage = weighted_coverage
                        best_combination = (row_s1, row_s2)

            # Combine the results into a single DataFrame
            result = pd.DataFrame([best_combination[0], best_combination[1]]) # type: ignore

            return max_coverage, result
     
        def get_best_combined_s1_s2_coverage(group: pd.DataFrame) -> Tuple[float, pd.DataFrame]:
            """
            Find the combined S1+S2 target with the highest weighted coverage
            :param group: The group of data to analyze
            :return: The row with the highest combined coverage
            """
            ghg_s1 = self.c.COLS.BASEYEAR_GHG_S1
            ghg_s2 = self.c.COLS.BASEYEAR_GHG_S2
            valid_rows = (group[ghg_s1] + group[ghg_s2]) > 0
            filtered_group = group[valid_rows].copy()
            if not filtered_group.empty:
            # Add column weighted_coverage to the DataFrame
                filtered_group.loc[:, 'weighted_coverage'] = (
                    (filtered_group[self.c.COLS.COVERAGE_S1] * filtered_group[ghg_s1] + filtered_group[self.c.COLS.COVERAGE_S2] * filtered_group[ghg_s2])
                    / (filtered_group[ghg_s1] + filtered_group[ghg_s2])
                ) 
                # Find the maximum weighted coverage
                sorted_group = filtered_group.sort_values(by='weighted_coverage', ascending=False)
                max_coverage = sorted_group['weighted_coverage'].max()
                result = sorted_group[sorted_group['weighted_coverage'] == max_coverage].drop('weighted_coverage', axis=1)
            else:
                max_coverage = 0.0
                # If there is no ghg data we can't calculate the combined coverage
                # So we just return the first row
                result = group.iloc[[0]]
            return max_coverage, result
        
        def settle_tied_coverage(single_scope_s1_s2: pd.DataFrame, combined_s1s2: pd.DataFrame) -> Tuple[bool, pd.DataFrame]:
            """
            Settle a tie between two targets by selecting the one with the highest ETargetReference
            :param single_scope_s1_s2: A dataframe with two rows representing the S1 and S2 targets
            :param combined_s1s2: A dataframe with the S1+S2 target
            :return: The target with the highest ETargetReference
            """
            # get values of the target types
            single_type_1 = single_scope_s1_s2[self.c.COLS.TARGET_REFERENCE_NUMBER].iloc[0]
            single_type_2 = single_scope_s1_s2[self.c.COLS.TARGET_REFERENCE_NUMBER].iloc[1]
            combined_type = combined_s1s2[self.c.COLS.TARGET_REFERENCE_NUMBER].iloc[0]

            if single_type_1 == single_type_2:
                if single_type_1 > combined_type:
                    return False, single_scope_s1_s2
                elif single_type_1 < combined_type:
                     return True, combined_s1s2
                else: # If both boundary coverage and target type are equal, then we prefer single scope targets
                    return False, single_scope_s1_s2
            else:
                min_single = min(single_type_1, single_type_2)
                if combined_type <= min_single:
                    return True, combined_s1s2
                else:
                    return False, single_scope_s1_s2
                
        # Initialize a list to collect the best entries
        best_entries = []

        # Iterate over each group of company_id and time_frame
        for _, group in grouped_data:
            # Check if both Scope 1 and Scope 2 exist in the group
            s1_data = group[group[self.c.COLS.SCOPE] == EScope.S1]
            s2_data = group[group[self.c.COLS.SCOPE] == EScope.S2]
            combined_s1_s2 = group[group[self.c.COLS.SCOPE] == EScope.S1S2]

            if not s1_data.empty and not s2_data.empty and not combined_s1_s2.empty:
                # If all three scopes exist, we need to compare the combined S1+S2 target with the individual S1 and S2 targets
                # Note that there may be more than one of each of the individual targets
                # Find the row in s1 and s2 with the highest coverage
                single_coverage, best_s1_s2_combination = get_best_s1_s2_combination(pd.concat([s1_data, s2_data]))
                # find the S1+S2 row with the highest weighted coverage
                combined_coverage, best_s1_s2_target = get_best_combined_s1_s2_coverage(combined_s1_s2)
            
                if single_coverage > combined_coverage:
                    best = best_s1_s2_combination
                    combined_best = False
                elif single_coverage < combined_coverage:
                    best = best_s1_s2_target
                    combined_best = True
                else:
                    # If it's a tie we prefer the target with the best target type 
                    combined_best, best = settle_tied_coverage(best_s1_s2_combination, best_s1_s2_target)            
                if combined_best:
                    best = self._split_s1s2_new(best)
                best['to_calculate'] = True
                best_entries.append(best)

            elif not s1_data.empty and not s2_data.empty and combined_s1_s2.empty:
                # If only S1 and S2 exist we find the combination of S1 and S2 with the highest combined coverage
                # Note that there may be more than one of each of the individual targets
                # Find the row in s1 and s2 with the highest coverage
                single_coverage, best_s1_s2_combination = get_best_s1_s2_combination(pd.concat([s1_data, s2_data]))
                #This must be the best combination, so we append it to the best entries
                best_s1_s2_combination['to_calculate'] = True
                best_entries.append(best_s1_s2_combination)

            elif not combined_s1_s2.empty:
                # If only the combined S1+S2 target exists, we find the S1+S2 target with the highest weighted coverage
                # Note that there may be more than one S1+S2 target
                _, best_s1_s2_target = get_best_combined_s1_s2_coverage(combined_s1_s2)
                #This must be the best combination, so we append it to the best entries after splitting
                # Process each row in best_s1_s2_target and apply _split_s1s2_new
                split_rows = []
                for _, row in best_s1_s2_target.iterrows():
                    split_rows.append(self._split_s1s2_new(pd.DataFrame([row])))

                # Concatenate the results
                best_s1_s2_target = pd.concat(split_rows, ignore_index=True)
                best_s1_s2_target['to_calculate'] = True
                best_entries.append(best_s1_s2_target)

            else:
                best_entries.append(group)

        # Concatenate all the best entries
        for df in best_entries:
            if isinstance(df, pd.DataFrame):
                # Restore the original column types for each dataframe in best_entries
                for col in df.columns:
                    if col in original_dtype:
                        df[col] = df[col].astype(original_dtype[col])
        
        result = pd.concat(best_entries, ignore_index=True)
        if isinstance(result, pd.Series):
            result = pd.DataFrame([result.values], columns = original_columns)
        else:
            result = result.reindex(columns=original_columns)
        result = result.set_index([self.c.COLS.COMPANY_ID, self.c.COLS.TIME_FRAME, self.c.COLS.SCOPE], drop=False)
       
        # Replace the identified rows in the original DataFrame with the sorted DataFrame
        remaining_data = data[~scope_1_2_mask]
        data = pd.concat([result, remaining_data], ignore_index=True)
        # Set the index again to maintain the original structure
       
        data = data.set_index([self.c.COLS.COMPANY_ID, self.c.COLS.TIME_FRAME, self.c.COLS.SCOPE], drop=False)
            
        return data



from enum import Enum
from typing import Optional, Tuple, Type, List, Any, cast

import pandas as pd
import numpy as np
import datetime
import os

from .interfaces import (
    ScenarioInterface,
    EScope,
    ETimeFrames,
    ETargetReference,
    Aggregation,
    AggregationContribution,
    ScoreAggregation,
    ScoreAggregationScopes,
    ScoreAggregations,
    PortfolioCompany,
    S3Category,
)
from .portfolio_aggregation import PortfolioAggregation, PortfolioAggregationMethod
from .configs import TemperatureScoreConfig
from . import data, utils


class ScenarioType(Enum):
    """
    A scenario defines which scenario should be run.
    """

    TARGETS = 1
    APPROVED_TARGETS = 2
    HIGHEST_CONTRIBUTORS = 3
    HIGHEST_CONTRIBUTORS_APPROVED = 4

    @staticmethod
    def from_int(value) -> Optional["ScenarioType"]:
        value_map = {
            1: ScenarioType.TARGETS,
            2: ScenarioType.APPROVED_TARGETS,
            3: ScenarioType.HIGHEST_CONTRIBUTORS,
            4: ScenarioType.HIGHEST_CONTRIBUTORS_APPROVED,
        }
        return value_map.get(value, None)


class EngagementType(Enum):
    """
    An engagement type defines how the companies will be engaged.
    """

    SET_TARGETS = 1
    SET_SBTI_TARGETS = 2

    @staticmethod
    def from_int(value) -> "EngagementType":
        """
        Convert an integer to an engagement type.

        :param value: The value to convert
        :return:
        """
        value_map = {
            0: EngagementType.SET_TARGETS,
            1: EngagementType.SET_SBTI_TARGETS,
        }
        return value_map.get(value, EngagementType.SET_TARGETS)

    @staticmethod
    def from_string(value: Optional[str]) -> "EngagementType":
        """
        Convert a string to an engagement type.

        :param value: The value to convert
        :return:
        """
        if value is None:
            return EngagementType.SET_TARGETS

        value_map = {
            "SET_TARGETS": EngagementType.SET_TARGETS,
            "SET_SBTI_TARGETS": EngagementType.SET_SBTI_TARGETS,
        }
        return value_map.get(value.upper(), EngagementType.SET_TARGETS)


class Scenario:
    """
    A scenario defines the action the portfolio holder will take to improve its temperature score.
    """

    scenario_type: Optional[ScenarioType]
    engagement_type: EngagementType

    def get_score_cap(self) -> float:
        if self.engagement_type == EngagementType.SET_TARGETS:
            return 1.75
        elif (
            self.scenario_type == ScenarioType.APPROVED_TARGETS
            or self.engagement_type == EngagementType.SET_SBTI_TARGETS
        ):
            return 1.5
        else:
            return np.NaN

    def get_default_score(self, default_score: float) -> float:
        if self.scenario_type == ScenarioType.TARGETS:
            return 1.75
        else:
            return default_score

    @staticmethod
    def from_dict(scenario_values: dict) -> Optional["Scenario"]:
        """
        Convert a dictionary to a scenario. The dictionary should have the following keys:

        * number: The scenario type as an integer
        * engagement_type: The engagement type as a string

        :param scenario_values: The dictionary to convert
        :return: A scenario object matching the input values or None, if no scenario could be matched
        """
        scenario = Scenario()
        scenario.scenario_type = ScenarioType.from_int(
            scenario_values.get("number", -1)
        )
        scenario.engagement_type = EngagementType.from_string(
            scenario_values.get("engagement_type", "")
        )

        if scenario.scenario_type is not None:
            return scenario
        else:
            return None

    @staticmethod
    def from_interface(
        scenario_values: Optional[ScenarioInterface],
    ) -> Optional["Scenario"]:
        """
        Convert a scenario interface to a scenario.

        :param scenario_values: The interface model instance to convert
        :return: A scenario object matching the input values or None, if no scenario could be matched
        """
        if scenario_values is None:
            return None

        scenario = Scenario()
        scenario.scenario_type = ScenarioType.from_int(scenario_values.number)
        scenario.engagement_type = EngagementType.from_string(
            scenario_values.engagement_type
        )

        if scenario.scenario_type is not None:
            return scenario
        else:
            return None


class TemperatureScore(PortfolioAggregation):
    """
    This class provides a temperature score based on the climate goals.

    :param default_score: The temp score if a company is not found
    :param model: The regression model to use
    :param config: A class defining the constants that are used throughout this class. This parameter is only required
                    if you'd like to overwrite a constant. This can be done by extending the TemperatureScoreConfig
                    class and overwriting one of the parameters.
    """

    def __init__(
        self,
        time_frames: List[ETimeFrames],
        scopes: List[EScope],
        default_score: float = TemperatureScoreConfig.DEFAULT_SCORE,
        model: int = TemperatureScoreConfig.MODEL_NUMBER,
        scenario: Optional[Scenario] = None,
        aggregation_method: PortfolioAggregationMethod = PortfolioAggregationMethod.WATS,
        grouping: Optional[List] = None,
        config: Type[TemperatureScoreConfig] = TemperatureScoreConfig,
    ):
        super().__init__(config)
        self.model = model
        self.c: Type[TemperatureScoreConfig] = config
        self.scenario: Optional[Scenario] = scenario
        self.default_score = default_score

        self.time_frames = time_frames
        self.scopes = scopes

        if self.scenario is not None:
            self.default_score = self.scenario.get_default_score(self.default_score)

        self.aggregation_method: PortfolioAggregationMethod = aggregation_method
        self.grouping: list = []
        if grouping is not None:
            self.grouping = grouping

        self.regression_model = pd.read_json(self.c.JSON_REGRESSION_MODEL)
        self.s3_calculation_test = TemperatureScoreConfig.TEST_S3_CALCULATION
        # Save code if we choose to use several models:
        # self.regression_model = self.regression_model[
        #     self.regression_model[self.c.COLS.MODEL] == self.model
        # ]

    def get_target_mapping(self, target: pd.Series) -> Optional[str]:
        """
        Map the target onto an AR6 target (None if not available).

        :param target: The target as a row of a dataframe
        :return: The mapped AR6 target
        """
        if (target[self.c.COLS.SCOPE] == EScope.S1S2
            or target[self.c.COLS.SCOPE] == EScope.S1S2S3
            ):
            map_scope = EScope.S1
        else:
            map_scope = target[self.c.COLS.SCOPE]
        if (
            target[self.c.COLS.TARGET_REFERENCE_NUMBER]
            .strip()
            .lower()
            .startswith(self.c.VALUE_TARGET_REFERENCE_INTENSITY_BASE)
        ):
            return self.c.INTENSITY_MAPPINGS.get(
                (target[self.c.COLS.INTENSITY_METRIC], map_scope), None
            )
        else:
            # Only first 3 characters of ISIC code are relevant for the absolute mappings
            return self.c.ABSOLUTE_MAPPINGS.get(
                (target[self.c.COLS.COMPANY_ISIC][:3], map_scope),
                self.c.ABSOLUTE_MAPPINGS.get(("other", map_scope)),
            )

    def get_annual_reduction_rate(self, target: pd.Series) -> Optional[float]:
        """
        Get the annual reduction rate (or None if not available).
        From version 1.5 the annual reduction rate is calculated as a 
        compund annual reduction rate, CAR.

        :param target: The target as a row of a dataframe
        :return: The annual reduction
        """
        check = pd.isnull(target[self.c.COLS.REDUCTION_AMBITION])
        check = check or pd.isnull(target[self.c.COLS.END_YEAR])
        check = check or pd.isnull(target[self.c.COLS.BASE_YEAR])
        check = check or (target[self.c.COLS.END_YEAR] <= target[self.c.COLS.BASE_YEAR])
        
        if check:
            return None
        elif target[self.c.COLS.REDUCTION_AMBITION] >= 1.0:
            return 1.0
        else:
            CAR = (1-target[self.c.COLS.REDUCTION_AMBITION]) ** float(
                1 / (target[self.c.COLS.END_YEAR] - target[self.c.COLS.BASE_YEAR])
            ) -1
            return CAR

    def get_regression(
        self, target: pd.Series
    ) -> Tuple[Optional[float], Optional[float]]:
        """
        Get the regression parameter and intercept from the model's output.

        :param target: The target as a row of a dataframe
        :return: The regression parameter and intercept
        """
        if pd.isnull(target[self.c.COLS.AR6]):
            return None, None

        regression = self.regression_model[
            (self.regression_model[self.c.COLS.VARIABLE] == target[self.c.COLS.AR6])
            & (
                self.regression_model[self.c.COLS.SLOPE]
                == self.c.SLOPE_MAP[target[self.c.COLS.TIME_FRAME]]
            )
        ]
        if len(regression) == 0:
            return None, None
        elif len(regression) > 1:
            # There should never be more than one potential mapping
            raise ValueError(
                "There is more than one potential regression parameter for this AR6 goal."
            )
        else:
            return (
                regression.iloc[0][self.c.COLS.PARAM],
                regression.iloc[0][self.c.COLS.INTERCEPT],
            )

    def _merge_regression(self, data: pd.DataFrame):
        """
        Merge the data with the regression parameters from the CDP-WWF Warming Function.
        :param data: The data to merge
        :return: The data set, amended with the regression parameters
        """     
        def get_slope(row: pd.Series) -> Optional[str]:
            return self.c.SLOPE_MAP.get(row[self.c.COLS.TIME_FRAME], None)

        data[self.c.COLS.SLOPE] = data.apply(lambda row: get_slope(row), axis=1) # type: ignore

        return pd.merge(
            left=data,
            right=self.regression_model,
            left_on=[self.c.COLS.SLOPE, self.c.COLS.AR6],
            right_on=[self.c.COLS.SLOPE, self.c.COLS.VARIABLE],
            how="left",
        )

    def get_score(self, target: pd.Series) -> Tuple[float, float]:
        """
        Get the temperature score for a certain target based on the annual reduction rate and the regression parameters.

        :param target: The target as a row of a data frame
        :return: The temperature score
        """
   
        if not target.to_calculate:
            return self.default_score, 1.0
        
        # CAR formula won't accept reduction of 100%, so assign the floor temperature score
        if abs(target[self.c.COLS.REDUCTION_AMBITION] - 1.0) < self.c.EPSILON:
            ts = self.c.TEMPERATURE_FLOOR
        # If target is set using CDP-WWF method, use the T_score from equation 5 in method doc
        elif target[self.c.COLS.TARGET_TYPE_AR6].lower() == "t_score":
            ts = max(
                target[self.c.COLS.BASE_YEAR_TS] 
                - (2040 - target[self.c.COLS.BASE_YEAR])
                * ((target[self.c.COLS.BASE_YEAR_TS] - target[self.c.COLS.END_YEAR_TS])
                   / (target[self.c.COLS.END_YEAR] - target[self.c.COLS.BASE_YEAR])),
                self.c.TEMPERATURE_FLOOR
            )
        else:    
            try:       
                ts = max(
                    target[self.c.COLS.REGRESSION_PARAM]
                    * target[self.c.COLS.ANNUAL_REDUCTION_RATE]
                    * 100 * -1      # According to the method doc 
                    + target[self.c.COLS.REGRESSION_INTERCEPT],
                    self.c.TEMPERATURE_FLOOR,
                )
            except TypeError as e:
                print(f"TypeError: {e}, {target[self.c.COLS.REGRESSION_PARAM]}, ambition: {target[self.c.COLS.REDUCTION_AMBITION]}")
                ts = self.default_score
        if target[self.c.COLS.SBTI_VALIDATED]:
            return ts, 0
        else:
            return (
                ts * self.c.SBTI_FACTOR
                + self.default_score * (1 - self.c.SBTI_FACTOR),
                0,
            )

    def aggregate_company_score(
        self, row: pd.Series, company_data: pd.DataFrame
    ) -> Tuple[float, float, list]:
        """
        Get the aggregated temperature score and a temperature result, which indicates how much of the score is 
        based on the default score for a certain company based on the emissions of company.

        :param company_data: The original data, grouped by company, time frame and scope category
        :param row: The row to calculate the temperature score for (if the scope of the row isn't s1s2s3, 
        it will return the original score
        :return: The aggregated temperature score for a company
        """

        if row[self.c.COLS.SCOPE] != EScope.S1S2S3 or (
            row[self.c.COLS.SCOPE] == EScope.S1S2S3
            and row[self.c.COLS.TARGET_TYPE_AR6].lower() == "t_score"
        ):
            return (
                row[self.c.COLS.TEMPERATURE_SCORE], 
                row[self.c.TEMPERATURE_RESULTS],
                row[self.c.COLS.TARGET_IDS],
            )
        
        s1 = company_data.xs(
            (row[self.c.COLS.COMPANY_ID], row[self.c.COLS.TIME_FRAME], EScope.S1)
        )
        s2 = company_data.xs(
            (row[self.c.COLS.COMPANY_ID], row[self.c.COLS.TIME_FRAME], EScope.S2)
        )
        s1s2 = company_data.xs(
            (row[self.c.COLS.COMPANY_ID], row[self.c.COLS.TIME_FRAME], EScope.S1S2)
        )
        s3 = company_data.xs(
            (row[self.c.COLS.COMPANY_ID], row[self.c.COLS.TIME_FRAME], EScope.S3)
        )

        s1_ghg = s1[self.c.COLS.GHG_SCOPE1]
        s2_ghg = s2[self.c.COLS.GHG_SCOPE2]
        s1s2_ghg = s1s2[self.c.COLS.GHG_SCOPE12]
        s3_ghg = s3[self.c.COLS.GHG_SCOPE3]

        if s1_ghg > 0 and s2_ghg > 0 and s3_ghg > 0:
            combined_TS = (s1.temperature_score * s1_ghg + 
            s2.temperature_score * s2_ghg + 
            s3.temperature_score * s3_ghg
            ) / (s1_ghg + s2_ghg + s3_ghg)
            combined_TR = (s1.temperature_results * s1_ghg +
            s2.temperature_results * s2_ghg +
            s3.temperature_results * s3_ghg
            ) / (s1_ghg + s2_ghg + s3_ghg)
            combined_targets = list((s1.target_ids or []) + 
            (s2.target_ids or []) + (s3.target_ids or [])) or []
        elif s1s2_ghg > 0 and s3_ghg > 0:
            if s1s2.to_calculate:
                combined_TS = (s1s2.temperature_score * s1s2_ghg + 
                s3.temperature_score * s3_ghg
                ) / (s1s2_ghg + s3_ghg)
            else:
                combined_TS = max(s1s2.temperature_score.item(), s3.temperature_score.item())

            combined_TR = (s1s2.temperature_results * s1s2_ghg + 
            s3.temperature_results * s3_ghg
            ) / (s1s2_ghg + s3_ghg)
            combined_targets = list((s1s2.target_ids or []) + 
            (s3.target_ids or [])) or []
        else:
            combined_TS = max(s1s2.temperature_score.item(), s3.temperature_score.item())
            combined_TR = row[self.c.TEMPERATURE_RESULTS]
            combined_targets = row[self.c.COLS.TARGET_IDS]
       

        return (
            combined_TS,
            combined_TR,
            combined_targets
        ) # type: ignore Pylance doesn't understand that TS and TR are floats
     

    def get_default_score(self, target: pd.Series) -> int:
        """
        Get the temperature score for a certain target based on the annual reduction rate and the regression parameters.

        :param target: The target as a row of a dataframe
        :return: The temperature score
        """
        if (
            pd.isnull(target[self.c.COLS.REGRESSION_PARAM])
            or pd.isnull(target[self.c.COLS.REGRESSION_INTERCEPT])
            or pd.isnull(target[self.c.COLS.ANNUAL_REDUCTION_RATE])
        ):
            return 1
        return 0

    def _prepare_data(self, data: pd.DataFrame):
        """
        Prepare the data such that it can be used to calculate the temperature score.

        :param data: The original data set as a pandas data frame
        :return: The extended data frame
        """
        # IDataProvider provides this optional field as an empty list by default
        # but if call TemperatureScore class with DataFrame directly then it may not be present
        if self.c.COLS.TARGET_IDS not in data.columns:
            data[self.c.COLS.TARGET_IDS] = [[] for _ in range(len(data))]

        # If scope S1S2S3 is in the list of scopes to calculate, we need to calculate the other two as well
        scopes = self.scopes.copy()
        if EScope.S1S2S3 in self.scopes and EScope.S1S2 not in self.scopes:
            scopes.append(EScope.S1S2)
        if EScope.S1S2S3 in scopes and EScope.S3 not in scopes:
            scopes.append(EScope.S3)
        # We also need to calculate scope 1 and scope 2 scores as of version 1.5
        if EScope.S1S2 in scopes and EScope.S1 not in scopes:
            scopes.append(EScope.S1)
            scopes.append(EScope.S2)

        data = data[
            data[self.c.COLS.SCOPE].isin(scopes)
            & data[self.c.COLS.TIME_FRAME].isin(self.time_frames)
        ].copy()

        data[self.c.COLS.TARGET_REFERENCE_NUMBER] = data[
            self.c.COLS.TARGET_REFERENCE_NUMBER
        #].replace({np.nan: self.c.VALUE_TARGET_REFERENCE_ABSOLUTE})
        ].replace({np.nan: ETargetReference.ABSOLUTE.value})
       
        # Replacing NaN with empty list in column target_ids
        data[self.c.COLS.TARGET_IDS] = data[self.c.COLS.TARGET_IDS].apply(
            lambda x: x if isinstance(x, list) else []
        )
        data[self.c.COLS.AR6] = data.apply(
            lambda row: self.get_target_mapping(row), axis=1 # type: ignore
        )  # type: ignore
        data[self.c.COLS.ANNUAL_REDUCTION_RATE] = data.apply(
            lambda row: self.get_annual_reduction_rate(row), axis=1 # type: ignore
        ) # type: ignore
        data = self._merge_regression(data)

        data[self.c.COLS.TEMPERATURE_SCORE], data[self.c.TEMPERATURE_RESULTS] = zip(
            *data.apply(lambda row: self.get_score(row), axis=1)
        )

        data = self.cap_scores(data)
        return data

    def _calculate_company_score(self, data):
        """
        Calculate the combined s1s2s3 scores for all companies.

        :param data: The original data set as a pandas data frame
        :return: The data frame, with an updated s1s2s3 temperature score
        """
     
        company_data = (
            data[
                [
                    self.c.COLS.COMPANY_ID,
                    self.c.COLS.TIME_FRAME,
                    self.c.COLS.SCOPE,
                    self.c.COLS.GHG_SCOPE1,
                    self.c.COLS.GHG_SCOPE2,
                    self.c.COLS.GHG_SCOPE12,
                    self.c.COLS.GHG_SCOPE3,
                    self.c.COLS.TEMPERATURE_SCORE,
                    self.c.COLS.TARGET_IDS,
                    self.c.TEMPERATURE_RESULTS,
                    self.c.COLS.TO_CALCULATE,
                ]
            ]
            .groupby(
                [self.c.COLS.COMPANY_ID, self.c.COLS.TIME_FRAME, self.c.COLS.SCOPE]
            )
            .agg(
                # take the mean of numeric columns, list-append self.c.COLS.TARGET_IDS
                {
                    self.c.COLS.GHG_SCOPE1: "mean",
                    self.c.COLS.GHG_SCOPE2: "mean",
                    self.c.COLS.GHG_SCOPE12: "mean",
                    self.c.COLS.GHG_SCOPE3: "mean",
                    self.c.COLS.TEMPERATURE_SCORE: "mean",
                    self.c.TEMPERATURE_RESULTS: "mean",
                    self.c.COLS.TARGET_IDS: "sum",
                    self.c.COLS.TO_CALCULATE: "any",
                }
            )
        )

        # sum pandas aggregator returns 0 where all input targets were None
        company_data.loc[
            company_data[self.c.COLS.TARGET_IDS] == 0, self.c.COLS.TARGET_IDS
        ] = None

        (
            data[self.c.COLS.TEMPERATURE_SCORE],
            data[self.c.TEMPERATURE_RESULTS],
            data[self.c.COLS.TARGET_IDS],
        ) = zip(
            *data.apply(
                lambda row: self.aggregate_company_score(row, company_data), axis=1
            )
        )
        return data

    def calculate(
        self,
        data: Optional[pd.DataFrame] = None,
        data_providers: Optional[List[data.DataProvider]] = None,
        portfolio: Optional[List[PortfolioCompany]] = None,
    ):
        """
        Calculate the temperature for a dataframe of company data. The columns in the data frame should be a combination
        of IDataProviderTarget and IDataProviderCompany.

        :param data: The data set (or None if the data should be retrieved)
        :param data_providers: A list of DataProvider instances. Optional, only required if data is empty.
        :param portfolio: A list of PortfolioCompany models. Optional, only required if data is empty.
        :return: A data frame containing all relevant information for the targets and companies
        """
        if data is None:
            if data_providers is not None and portfolio is not None:
                data = utils.get_data(data_providers, portfolio)
            else:
                raise ValueError(
                    "You need to pass and either a data set or a list of data providers and companies"
                )

        data = self._prepare_data(data)

        if EScope.S1S2 in self.scopes:
            data = self._calculate_s1s2_score(data)

        # data = self._aggregate_s3_score(data)
        if self.s3_calculation_test:
            s3_data_dump = data[data[self.c.COLS.SCOPE] == EScope.S3].copy()
            # get path to save the data
            path = os.path.join(os.path.dirname(__file__), "../examples/data/local/s3_data_dump.xlsx")
            s3_data_dump.to_excel(path, index=False)
        data = self._aggregate_s3_score(data)

        if EScope.S1S2S3 in self.scopes:
            data = self._calculate_company_score(data)

        # We need to filter the scopes again, because we might have had to add a scope in the preparation step
        data = data[data[self.c.COLS.SCOPE].isin(self.scopes)]
        data.drop(columns=['to_calculate'], inplace=True)
        
        return data

    def _get_aggregations(
        self, data: pd.DataFrame, total_companies: int
    ) -> Tuple[Aggregation, pd.Series, pd.Series]:
        """
        Get the aggregated score over a certain data set. Also calculate the (relative) contribution of each company

        :param data: A data set, containing one row per company
        :return: An aggregated score and the relative and absolute contribution of each company
        """
        data = data.copy()
        weighted_scores = self._calculate_aggregate_score(
            data, self.c.COLS.TEMPERATURE_SCORE, self.aggregation_method
        )
        data[self.c.COLS.CONTRIBUTION_RELATIVE] = weighted_scores / (
            weighted_scores.sum() / 100
        )
        data[self.c.COLS.CONTRIBUTION] = weighted_scores
        contributions = (
            data.sort_values(self.c.COLS.CONTRIBUTION_RELATIVE, ascending=False)
            .where(pd.notnull(data), None)
            .to_dict(orient="records")
        )
        aggregation = Aggregation(
            score=weighted_scores.sum(),
            proportion=len(weighted_scores) / (total_companies / 100.0),
            contributions=[
            AggregationContribution.model_validate(contribution)
            for contribution in contributions
            ],
        )
        return (
            aggregation,
            data[self.c.COLS.CONTRIBUTION_RELATIVE],
            data[self.c.COLS.CONTRIBUTION],
        )

    def _get_score_aggregation(
        self, data: pd.DataFrame, time_frame: ETimeFrames, scope: EScope
    ) -> Optional[ScoreAggregation]:
        """
        Get a score aggregation for a certain time frame and scope, for the data set as a whole and for the different
        groupings.

        :param data: The whole data set
        :param time_frame: A time frame
        :param scope: A scope
        :return: A score aggregation, containing the aggregations for the whole data set and each individual group
        """
        filtered_data = data[
            (data[self.c.COLS.TIME_FRAME] == time_frame)
            & (data[self.c.COLS.SCOPE] == scope)
        ].copy()
        filtered_data[self.grouping] = filtered_data[self.grouping].fillna("unknown")
        total_companies = len(filtered_data)
        if not filtered_data.empty:
            (
                score_aggregation_all,
                filtered_data[self.c.COLS.CONTRIBUTION_RELATIVE],
                filtered_data[self.c.COLS.CONTRIBUTION],
            ) = self._get_aggregations(filtered_data, total_companies)
            score_aggregation = ScoreAggregation(
                grouped={},
                all=score_aggregation_all,
                influence_percentage=self._calculate_aggregate_score(
                    filtered_data, self.c.TEMPERATURE_RESULTS, self.aggregation_method
                ).sum()
                * 100,
            )

            # If there are grouping column(s) we'll group in pandas and pass the results to the aggregation
            if len(self.grouping) > 0:
                grouped_data = filtered_data.groupby(self.grouping)
                for group_names, group in grouped_data:
                    group_name_joined = (
                        group_names
                        if type(group_names) == str
                        else "-".join([str(group_name) for group_name in group_names])
                    )
                    (
                        score_aggregation.grouped[group_name_joined],
                        _,
                        _,
                    ) = self._get_aggregations(group.copy(), total_companies)
            return score_aggregation
        else:
            return None

    def aggregate_scores(self, data: pd.DataFrame) -> ScoreAggregations:
        """
        Aggregate scores to create a portfolio score per time_frame (short, mid, long).

        :param data: The results of the calculate method
        :return: A weighted temperature score for the portfolio
        """

        score_aggregations = ScoreAggregations() # type: ignore
        for time_frame in self.time_frames:
            score_aggregation_scopes = ScoreAggregationScopes() # type: ignore
            for scope in self.scopes:
                score_aggregation_scopes.__setattr__(
                    scope.name, self._get_score_aggregation(data, time_frame, scope)
                )
            score_aggregations.__setattr__(time_frame.value, score_aggregation_scopes)

        return score_aggregations

    def cap_scores(self, scores: pd.DataFrame) -> pd.DataFrame:
        """
        Cap the temperature scores in the input data frame to a certain value, based on the scenario that's being used.
        This can either be for the whole data set, or only for the top X contributors.

        :param scores: The data set with the temperature scores
        :return: The input data frame, with capped scores
        """
        if self.scenario is None:
            return scores
              
        if self.scenario.scenario_type == ScenarioType.APPROVED_TARGETS:
            score_based_on_target = ~pd.isnull(
                scores[self.c.COLS.TARGET_REFERENCE_NUMBER]
            )
            scores.loc[
                score_based_on_target, self.c.COLS.TEMPERATURE_SCORE
            ] = scores.loc[score_based_on_target, self.c.COLS.TEMPERATURE_SCORE].apply(
                lambda x: min(x, self.scenario.get_score_cap()) # type: ignore
            )
        elif self.scenario.scenario_type == ScenarioType.HIGHEST_CONTRIBUTORS:
            # Cap scores of 10 highest contributors per time frame-scope combination
     
            aggregations = self.aggregate_scores(scores)
            for time_frame in self.time_frames:
                for scope in self.scopes:
                    number_top_contributors = min(
                        10,
                        len(
                            aggregations[time_frame.value][scope.name].all.contributions
                        ),
                    )
                    for contributor in range(number_top_contributors):
                        company_name = aggregations[time_frame.value][
                            scope.name
                        ].all.contributions[contributor][self.c.COLS.COMPANY_NAME]
                        company_mask = (
                            (scores[self.c.COLS.COMPANY_NAME] == company_name)
                            & (scores[self.c.COLS.SCOPE] == scope)
                            & (scores[self.c.COLS.TIME_FRAME] == time_frame)
                        )
                        scores.loc[
                            company_mask, self.c.COLS.TEMPERATURE_SCORE
                        ] = scores.loc[
                            company_mask, self.c.COLS.TEMPERATURE_SCORE
                        ].apply( # type: ignore
                            lambda x: min(x, self.scenario.get_score_cap()) # type: ignore
                        )
        elif self.scenario.scenario_type == ScenarioType.HIGHEST_CONTRIBUTORS_APPROVED:
            score_based_on_target = scores[self.c.COLS.ENGAGEMENT_TARGET]
            scores.loc[
                score_based_on_target, self.c.COLS.TEMPERATURE_SCORE
            ] = scores.loc[score_based_on_target, self.c.COLS.TEMPERATURE_SCORE].apply(
                lambda x: min(x, self.scenario.get_score_cap()) # type: ignore
            )
        return scores

    def anonymize_data_dump(self, scores: pd.DataFrame) -> pd.DataFrame:
        """
        Anonymize the scores by deleting the company IDs, ISIN and renaming the companies .

        :param scores: The data set with the temperature scores
        :return: The input data frame, anonymized
        """
        scores.drop(
            columns=[self.c.COLS.COMPANY_ID, self.c.COLS.COMPANY_ISIN], inplace=True
        )
        for index, company_name in enumerate(scores[self.c.COLS.COMPANY_NAME].unique()):
            scores.loc[
                scores[self.c.COLS.COMPANY_NAME] == company_name,
                self.c.COLS.COMPANY_NAME,
            ] = "Company" + str(index + 1)
        return scores

    def _calculate_s1s2_score(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Calculate the combined S1S2 score each combination of company_id, time_frame and scope.
        If it was possible på calculate individual S1 and S2 scores, then we calculate the 
        aggregated S1S2 score.

        :param data: The data to calculate the S1S2 score for.
        :return: The S1S2 score.
        """
              
        s1s2_data_mask = data[self.c.COLS.SCOPE].isin([EScope.S1, EScope.S2, EScope.S1S2])
        s1s2_data = data[s1s2_data_mask].copy()
        grouped_data = s1s2_data.groupby(['company_id', 'time_frame'])

        for _, group in grouped_data:
            s1s2_score = np.nan # Initailize the s1s2_score to NaN
            s1_s2_results = 1.0 # Initialize the s1_s2_results to 1.0
            if (group.loc[group['scope'] == EScope.S1,'to_calculate'].any() or
                group.loc[group['scope'] == EScope.S2, 'to_calculate'].any()
            ):
                # S1 and S2 scores are either calulted or default scores from earlier
                s1_score = float(group.loc[group['scope'] == EScope.S1, 'temperature_score'].item())
                s2_score = float(group.loc[group['scope'] == EScope.S2, 'temperature_score'].item())
                s1_results = float(group.loc[group['scope'] == EScope.S1, 'temperature_results'].item())
                s2_results = float(group.loc[group['scope'] == EScope.S2, 'temperature_results'].item())
                s1_ghg = float(group.loc[group['scope'] == EScope.S1, 'ghg_s1'].item())
                s2_ghg = float(group.loc[group['scope'] == EScope.S2, 'ghg_s2'].item())
                # REQUIREMENT 6.5 Temperature Score Aggregation
                
                if not np.isnan(s1_ghg) and not np.isnan(s2_ghg) and s1_ghg != 0 and s2_ghg != 0:
                    try:
                        s1s2_score = (s1_score * s1_ghg + s2_score * s2_ghg) / (s1_ghg + s2_ghg)
                        s1_s2_results = (s1_results * s1_ghg + s2_results * s2_ghg) / (s1_ghg + s2_ghg)
                    except ZeroDivisionError:
                       print("Division by zero")
                       print(f"Target ID {group['target_ids'].values[0]}")
                else:
                    s1s2_score = max(s1_score, s2_score)
                    s1_s2_results = 0.5
                group.loc[group['scope'] == EScope.S1S2, 'temperature_score'] = s1s2_score
                group.loc[group['scope'] == EScope.S1S2, 'temperature_results'] = s1_s2_results
                group.loc[group['scope'] == EScope.S1S2, 'target_type'] = (
                    group.loc[group['scope'] == EScope.S1, 'target_type'].values[0]
                )
                s1_target_ids = group[group['scope'] == EScope.S1]['target_ids'].values 
                s2_target_ids = group[group['scope'] == EScope.S2]['target_ids'].values
                combined_target_ids = list(set().union(*s1_target_ids, *s2_target_ids))
                idx = group.index[group['scope'] == EScope.S1S2][0] 
                group.at[idx, 'target_ids'] = combined_target_ids
                        
                # group = group.infer_objects()
               
                # data.update(group) Works but throws a FutureWarning
               
                group = group.copy() # Important!
                cols_to_update = group.columns.tolist() # Or group.columns if no potential for missing columns
                for index in group.index:
                    if index in data.index: # Essential check to avoid KeyError.
                        data.loc[index, cols_to_update] = group.loc[index, cols_to_update].values #.values is important
   
        return data
          
    def _aggregate_s3_score(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Aggregate Scope 3 category scores into a single score per company and time frame.

        Steps:
        1) Calculate the mean temperature score for `s3_category == CAT_15`.
        2) Remove original CAT_15 rows and append the mean CAT_15 score back to the data.
        3) For each company and time frame:
        a) If GHG data is available for all categories, calculate a weighted average.
        b) If not, calculate a simple average of the scores.
        4) Replace the original scores with the aggregated score.
        """
        # Separate Scope 3 data
        s3_data = data[data['scope'] == EScope.S3].copy()

        # If there's no s3_category column or no S3 data, nothing to aggregate
        if 's3_category' not in s3_data.columns or s3_data.empty:
            return data

        # Step 1: Calculate mean temperature score for CAT_15
        # cat_15_data = s3_data[s3_data['s3_category'] == S3Category.CAT_15].copy()
        # mean_cat_15 = cat_15_data.groupby(['company_id', 'time_frame']).agg({
        #     'temperature_score': 'mean',
        #     'target_ids': lambda x: list(set().union(*x)),
        #     'target_type': 'first',
        #     'scope': 'first',
        #     's3_category': 'first',
        #     'ghg_s3_15': 'first',
        #     'company_name': 'first'
        # }).reset_index()
        # Step 1: Calculate mean temperature score for CAT_15
        cat_15_data = s3_data[s3_data['s3_category'] == S3Category.CAT_15].copy()

        # Define the aggregation functions for special columns
        agg_funcs = {
            'temperature_score': 'mean',
            'target_ids': lambda x: list(set().union(*x))
        }

        # Specify group by columns
        group_cols = ['company_id', 'time_frame']

        # Get all other columns that need to use 'first'
        other_cols = [col for col in cat_15_data.columns if col not in group_cols + list(agg_funcs.keys())]

        # Set 'first' as the aggregation function for other columns
        for col in other_cols:
            agg_funcs[col] = 'first'

        # Perform the aggregation
        mean_cat_15 = cat_15_data.groupby(group_cols).agg(agg_funcs).reset_index()


        # Step 2: Remove original CAT_15 rows and append the mean score
        s3_data = s3_data[s3_data['s3_category'] != S3Category.CAT_15]
        s3_data = pd.concat([s3_data, mean_cat_15], ignore_index=True)

        # Ensure all required columns are present
        required_columns = data.columns.tolist()
        for col in required_columns:
            if col not in s3_data.columns:
                s3_data[col] = np.nan
        s3_data = s3_data[required_columns]

        # Step 3: Aggregate scores for each company and time frame
        aggregated_rows = []
        grouped = s3_data.groupby(['company_id', 'time_frame'])
        for _ , group in grouped:
            temperature_scores = np.full(15, 3.4) # Set all scores to the default value
            ghg_columns = [f'ghg_s3_{i}' for i in range(1, 16)]
            ghg_values = group.iloc[0][ghg_columns].values.astype(float)
           
            for _, row in group.iterrows():
                s3_category = row['s3_category']
                if pd.notna(s3_category) and isinstance(s3_category, S3Category):
                    index = s3_category.value - 1  # Convert category to index
                    temperature_scores[index] = row['temperature_score']

            # Calculate weighted or simple average
            if not np.isnan(ghg_values).any().all() and np.nansum(ghg_values) > 0.0:
                weighted_avg = np.sum(ghg_values * temperature_scores) / np.sum(ghg_values)
            else:
                weighted_avg = temperature_scores.mean()

            # Create aggregated row
            aggregated_row = group.iloc[0].copy()
            aggregated_row['temperature_score'] = weighted_avg
            aggregated_row['target_ids'] = sum(group['target_ids'], [])
            if S3Category.CAT_H_LINE in group['s3_category'].values:
                aggregated_row['s3_category'] = S3Category.CAT_H_LINE
            else:
                aggregated_row['s3_category'] = 'Aggregated'

            aggregated_rows.append(aggregated_row)

        # Step 4: Replace original s3_data with aggregated data
        aggregated_data = pd.DataFrame(aggregated_rows)
        data = data[data['scope'] != EScope.S3]
        data = pd.concat([data, aggregated_data], ignore_index=True)
        data.sort_values(by=['company_id', 'time_frame', 'scope'], inplace=True)

        return data


"""
This file defines the constants used throughout the different classes. In order to redefine these settings whilst using
the module, extend the respective config class and pass it to the class as the "constants" parameter.
"""
import os

from ITR.interfaces import ETimeFrames, EScope, S3Category, ETargetReference


class ColumnsConfig:
    # Define a constant for each column used in the input data
    COMPANY_ID = "company_id"
    COMPANY_ISIN = "company_isin"
    COMPANY_LEI = "company_lei"
    COMPANY_ISIC = "isic"
    TARGET_IDS = "target_ids"
    REGRESSION_PARAM = "param"
    REGRESSION_INTERCEPT = "intercept"
    MARKET_CAP = "company_market_cap"
    INVESTMENT_VALUE = "investment_value"
    COMPANY_ENTERPRISE_VALUE = "company_enterprise_value"
    COMPANY_EV_PLUS_CASH = "company_ev_plus_cash"
    COMPANY_TOTAL_ASSETS = "company_total_assets"
    TARGET_REFERENCE_NUMBER = "target_type"
    SCOPE = "scope"
    SCOPE3_CATEGORY = "s3_category"
    AR6 = "ar6"
    REDUCTION_FROM_BASE_YEAR = "reduction_from_base_year"
    START_YEAR = "start_year"
    VARIABLE = "variable"
    SLOPE = "slope"
    TIME_FRAME = "time_frame"
    MODEL = "model"
    ANNUAL_REDUCTION_RATE = "annual_reduction_rate"
    EMISSIONS_IN_SCOPE = "emissions_in_scope"
    TEMPERATURE_SCORE = "temperature_score"
    COMPANY_NAME = "company_name"
    OWNED_EMISSIONS = "owned_emissions"
    COUNTRY = "country"
    SECTOR = "sector"
    GHG_SCOPE1 = "ghg_s1"
    GHG_SCOPE2 = "ghg_s2"
    GHG_SCOPE12 = "ghg_s1s2"
    GHG_SCOPE3 = "ghg_s3"
    COMPANY_REVENUE = "company_revenue"
    CASH_EQUIVALENTS = "company_cash_equivalents"
    TARGET_CLASSIFICATION = "target_classification"
    REDUCTION_AMBITION = "reduction_ambition"
    BASE_YEAR = "base_year"
    END_YEAR = "end_year"
    SBTI_VALIDATED = "sbti_validated"
    TARGET_CONFIRM_DATE = "statement_date"
    ACHIEVED_EMISSIONS = "achieved_reduction"
    ISIC = "isic"
    INDUSTRY_LVL1 = "industry_level_1"
    INDUSTRY_LVL2 = "industry_level_2"
    INDUSTRY_LVL3 = "industry_level_3"
    INDUSTRY_LVL4 = "industry_level_4"
    COVERAGE_S1 = "coverage_s1"
    COVERAGE_S2 = "coverage_s2"
    COVERAGE_S3 = "coverage_s3"
    INTENSITY_METRIC = "intensity_metric"
    INTENSITY_METRIC_AR6 = "intensity_metric"
    TARGET_TYPE_AR6 = "target_type"
    AR6_VARIABLE = "ar6_variable"
    REGRESSION_MODEL = "Regression_model"
    BASEYEAR_GHG_S1 = "base_year_ghg_s1"
    BASEYEAR_GHG_S2 = "base_year_ghg_s2"
    BASEYEAR_GHG_S1S2 = "base_year_ghg_s1s2"
    BASEYEAR_GHG_S3 = "base_year_ghg_s3"
    REGION = "region"
    ENGAGEMENT_TARGET = "engagement_target"
    BASE_YEAR_TS = "base_year_ts"
    END_YEAR_TS = "end_year_ts"
    TO_CALCULATE = "to_calculate"

    # Scope 3 categories - from fundamental data
    GHG_S3_1 = "ghg_s3_1"
    GHG_S3_2 = "ghg_s3_2"
    GHG_S3_3 = "ghg_s3_3"
    GHG_S3_4 = "ghg_s3_4"
    GHG_S3_5 = "ghg_s3_5"
    GHG_S3_6 = "ghg_s3_6"
    GHG_S3_7 = "ghg_s3_7"
    GHG_S3_8 = "ghg_s3_8"
    GHG_S3_9 = "ghg_s3_9"
    GHG_S3_10 = "ghg_s3_10"
    GHG_S3_11 = "ghg_s3_11"
    GHG_S3_12 = "ghg_s3_12"
    GHG_S3_13 = "ghg_s3_13"
    GHG_S3_14 = "ghg_s3_14"
    GHG_S3_15 = "ghg_s3_15"

    # AR6 mapping columns
    PARAM = "param"
    INTERCEPT = "intercept"

    # Output columns
    WEIGHTED_TEMPERATURE_SCORE = "weighted_temperature_score"
    CONTRIBUTION_RELATIVE = "contribution_relative"
    CONTRIBUTION = "contribution"


class PortfolioAggregationConfig:
    COLS = ColumnsConfig


class TemperatureScoreConfig(PortfolioAggregationConfig):

    """
    This factor determines what part of the temperature for a not SBTi-validated company should be the TS and what part
    should be the default score.
    The calculated temperature score should not be lower than the current level of
    global warning which is expressed through the temperature floor constant.
    """

    SBTI_FACTOR = 1
    DEFAULT_SCORE: float = 3.4
    TEMPERATURE_FLOOR: float = 1.5
    TEST_S3_CALCULATION = False # Set to True to print S3 calculation results   
  
    JSON_REGRESSION_MODEL = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "inputs",
        "AR6_regression_model.json",
    )
    MODEL_NUMBER: int = 1
    DEFAULT_INDUSTRY = "Others"

    # VALUE_TARGET_REFERENCE_ABSOLUTE = "absolute"
    # VALUE_TARGET_REFERENCE_T_SCORE = "t_score"
    # VALUE_TARGET_REFERENCE_INTENSITY = "intensity"
    VALUE_TARGET_REFERENCE_INTENSITY_BASE = "inte"

    SLOPE_MAP = {
        ETimeFrames.SHORT: "slopeCA5",
        ETimeFrames.MID: "slopeCA10",
        ETimeFrames.LONG: "slopeCA30",
    }
    #test this
    VALUE_TARGET_REFERENCE = ETargetReference

    INTENSITY_MAPPINGS = {
        ("Revenue", EScope.S1): "INT.emKyoto_gdp",
        ("Revenue", EScope.S2): "INT.emCO2energysupply_SE",
        ("Revenue", EScope.S3): "INT.emKyoto_gdp",
        ("Product", EScope.S1): "INT.emKyoto_gdp",
        ("Product", EScope.S2): "INT.emCO2energysupply_SE",
        ("Product", EScope.S3): "INT.emKyoto_gdp",
        ("Cement", EScope.S1): "INT.emKyoto_gdp",
        ("Cement", EScope.S2): "INT.emCO2energysupply_SE",
        ("Cement", EScope.S3): "INT.emKyoto_gdp",
        ("Oil", EScope.S1): "INT.emKyoto_gdp",
        ("Oil", EScope.S2): "INT.emCO2energysupply_SE",
        ("Oil", EScope.S3): "INT.emKyoto_gdp",
        ("Steel", EScope.S1): "INT.emKyoto_gdp",
        ("Steel", EScope.S2): "INT.emCO2energysupply_SE",
        ("Steel", EScope.S3): "INT.emKyoto_gdp",
        ("Aluminum", EScope.S1): "INT.emKyoto_gdp",
        ("Aluminum", EScope.S2): "INT.emCO2energysupply_SE",
        ("Aluminum", EScope.S3): "INT.emKyoto_gdp",
        ("Power", EScope.S1): "INT.emCO2energysupply_SE",
        ("Power", EScope.S2): "INT.emCO2energysupply_SE",
        ("Power", EScope.S3): "INT.emKyoto_gdp",
        ("Other", EScope.S1): "INT.emKyoto_gdp",
        ("Other", EScope.S2): "INT.emCO2energysupply_SE",
        ("Other", EScope.S3): "INT.emKyoto_gdp",
    }
    intensity_metric_types = set(metric for metric, _ in INTENSITY_MAPPINGS.keys())

    ABSOLUTE_MAPPINGS = {
        # B06: Extraction Of Crude Petroleum And Natural Gas
        ("B06", EScope.S1): "Emissions|Kyoto Gases", 
        ("B06", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("B06", EScope.S3): "Emissions|Kyoto Gases",
        # C23: Manufacture Of Other Non-Metallic Mineral Products ie Cement
        ("C23", EScope.S1): "Emissions|CO2|Energy and Industrial Processes",
        ("C23", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("C23", EScope.S3): "Emissions|Kyoto Gases",
        # C24: Manufacture Of Basic Metals ie Steel, Aluminum
        ("C24", EScope.S1): "Emissions|CO2|Energy and Industrial Processes",
        ("C24", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("C24", EScope.S3): "Emissions|Kyoto Gases",
        # D35: Electricity, Gas, Steam And Air Conditioning Supply
        ("D35", EScope.S1): "Emissions|CO2|Energy|Supply",
        ("D35", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("D35", EScope.S3): "Emissions|Kyoto Gases",
        # H49: Land Transport And Transport Via Pipelines
        ("H49", EScope.S1): "Emissions|Kyoto Gases",
        ("H49", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("H49", EScope.S3): "Emissions|Kyoto Gases",
        # H50: Water Transport
        ("H50", EScope.S1): "Emissions|Kyoto Gases",
        ("H50", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("H50", EScope.S3): "Emissions|Kyoto Gases",
        # H51: Air Transport
        ("H51", EScope.S1): "Emissions|Kyoto Gases",
        ("H51", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("H51", EScope.S3): "Emissions|Kyoto Gases",
        # H52: Warehousing And Support Activities For Transportation
        ("H52", EScope.S1): "Emissions|Kyoto Gases",
        ("H52", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("H52", EScope.S3): "Emissions|Kyoto Gases",
        # H53: Postal And Courier Activities
        ("H53", EScope.S1): "Emissions|Kyoto Gases",
        ("H53", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("H53", EScope.S3): "Emissions|Kyoto Gases",
        ("other", EScope.S1): "Emissions|Kyoto Gases",
        ("other", EScope.S2): "Emissions|CO2|Energy|Supply",
        ("other", EScope.S3): "Emissions|Kyoto Gases",
    }

    TEMPERATURE_RESULTS = "temperature_results"
    INVESTMENT_VALUE = "investment_value"
    
    S3_CATEGORY_MAPPINGS = {
            S3Category.CAT_1: "ghg_s3_1",
            S3Category.CAT_2: "ghg_s3_2",
            S3Category.CAT_3: "ghg_s3_3",
            S3Category.CAT_4: "ghg_s3_4",
            S3Category.CAT_5: "ghg_s3_5",
            S3Category.CAT_6: "ghg_s3_6",
            S3Category.CAT_7: "ghg_s3_7",
            S3Category.CAT_8: "ghg_s3_8",
            S3Category.CAT_9: "ghg_s3_9",
            S3Category.CAT_10: "ghg_s3_10",
            S3Category.CAT_11: "ghg_s3_11",
            S3Category.CAT_12: "ghg_s3_12",
            S3Category.CAT_13: "ghg_s3_13",
            S3Category.CAT_14: "ghg_s3_14",
            S3Category.CAT_15: "ghg_s3_15",
            S3Category.CAT_H_LINE: "ghg_s3"
    }
    EPSILON = 1e-6

class PortfolioCoverageTVPConfig(PortfolioAggregationConfig):
    # Store CTA file in user's home directory to ensure fresh data and avoid permission issues
    # Falls back to bundled file if download fails (e.g., firewall restrictions)
    FILE_TARGETS = os.path.join(
        os.path.expanduser("~"),
        ".itr-tool",
        "current-Companies-Taking-Action.xlsx",
    )
    # To avoid CTA file being downloaded every time and use a local file instead, set USE_LOCAL_CTA = True and set the
    # path to the local file in FILE_TARGETS_CUSTOM_PATH
    FILE_TARGETS_CUSTOM_PATH = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "inputs",
        "current-Companies-Taking-Action.xlsx",
    )
    USE_LOCAL_CTA = False
    # If the CTA file is older than a week, the file will be downloaded again
    SKIP_CTA_FILE_IF_EXISTS = True
     # Temporary URL until the SBTi website is updated
    #CTA_FILE_URL = "https://docs.google.com/spreadsheets/d/e/2PACX-1vTn5UZIBmOxWKNFpmQGWWDczMvdBJ74l2_j0emUH9mxEKylHqh3oMLhu2FXtAV7-bqDxy9Yz_hkWzu8/pub?output=xlsx"
    CTA_FILE_URL = "https://sb-ti.files.svdcdn.com/staging/files/companies-excel.xlsx"
    USE_CUSTOM_FILE_TARGETS_PATH = False
    OUTPUT_TARGET_STATUS = "sbti_target_status"
    OUTPUT_WEIGHTED_TARGET_STATUS = "weighted_sbti_target_status"
    VALUE_STATUS_SET = "Targets set"
    VALUE_STATUS_COMMITTED = "Committed"
    VALUE_STATUS_REMOVED = "Commitment removed"
    
    # VALUE_TARGET_NO = "No target"
    # VALUE_TARGET_SET = "Near-term"
    # VALUE_ACTION_COMMITTED = "Commitment"
    # VALUE_ACTION_TARGET = "Target"

    TARGET_SCORE_MAP = {
        VALUE_STATUS_REMOVED: 0,
        VALUE_STATUS_COMMITTED: 0,
        VALUE_STATUS_SET: 100,
    }

    # SBTi targets overview (TVP coverage)
    COL_COMPANY_NAME = "company_name"
    COL_COMPANY_ISIN = "isin"
    COL_COMPANY_LEI = "lei"
    #COL_ACTION = "Action"
    COL_TARGET = "near_term_status"


from typing import Type, List
import logging
from pydantic import ValidationError
import pandas as pd
from ITR.data.data_provider import DataProvider
from ITR.interfaces import IDataProviderCompany, IDataProviderTarget, S3Category
from ITR.configs import ColumnsConfig

class_definitions = {
    "fundamental": {
        "company_name": str,
        "company_id": str,
        "isic": str,
        "country": str,
        "region": str,
        "industry_level_1": str,
        "industry_level_2": str,
        "industry_level_3": str,
        "industry_level_4": str,
        "sector": str,
        "ghg_s1s2": str,
        "ghg_s3": str,
        "company_revenue": str,
        "company_market_cap": str,
        "company_enterprise_value": str,
        "company_total_assets": str,
        "company_cash_equivalents": str,
    },
    "target": {
        "company_name": str,
        "company_id": str,
        "target_type": str,
        "intensity_metric": str,
        "scope": str,
        "coverage_s1": str,
        "coverage_s2": str,
        "coverage_s3": str,
        "reduction_ambition": str,
        "base_year": str,
        "end_year": str,
        "start_year": str,
        "base_year_ghg_s1": str,
        "base_year_ghg_s2": str,
        "base_year_ghg_s3": str,
        "achieved_reduction": str,
        "target_ids": str,
    },
}


class InMemoryProvider(DataProvider):
    """
    Data provider to read in-memory dict.

    :param fundamental: A dictionary with the fundamental data
    :param targets: A dictionary with the target data
    """

    def __init__(
        self,
        fundamental: class_definitions["fundamental"],
        targets: class_definitions["target"],
        config: Type[ColumnsConfig] = ColumnsConfig,
    ):
        super().__init__()
        self.data_fundamental = pd.DataFrame(fundamental)
        self.data_targets = pd.DataFrame(targets)
        self.c = config

    def get_targets(self, company_ids: list) -> List[IDataProviderTarget]:
        """
        Get all relevant targets for a list of company ids (ISIN). This method should return a list of
        IDataProviderTarget instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the targets
        """
        model_targets = self._target_df_to_model(self.data_targets)
        model_targets = [
            target for target in model_targets if target.company_id in company_ids
        ]
        return model_targets

    def _target_df_to_model(self, df_targets):
        """
        transforms target Dataframe into list of IDataProviderTarget instances
        
        :param df_targets: pandas Dataframe with targets
        :return: A list containing the targets
        """
        logger = logging.getLogger(__name__)
        targets = df_targets.to_dict(orient="records")
        model_targets: List[IDataProviderTarget] = []

        for target in targets:
            try:
                # Replace NaN/NaT with None so Pydantic handles Optional fields correctly
                cleaned = {}
                for k, v in target.items():
                    try:
                        if v is None or (not isinstance(v, (list, dict, str)) and pd.isna(v)):
                            cleaned[k] = None
                        else:
                            cleaned[k] = v
                    except (ValueError, TypeError):
                        cleaned[k] = v
                target = cleaned
                # Convert float years to int (Excel reads int columns as float)
                for year_col in ['base_year', 'end_year', 'start_year']:
                    if year_col in target and target[year_col] is not None:
                        target[year_col] = int(target[year_col])
                # Map s3_category if present
                if 's3_category' in target and target['s3_category'] is not None:
                    target['s3_category'] = S3Category(int(target['s3_category']))
                model_targets.append(IDataProviderTarget.model_validate(target))
            except (ValidationError, TypeError) as e:
                logger.warning(
                    f"Target validation failed and will be skipped: {e}"
                )
                continue

        return model_targets

    def get_company_data(self, company_ids: list) -> List[IDataProviderCompany]:
        """
        Get all relevant data for a list of company ids (ISIN). This method should return a list of IDataProviderCompany
        instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the company data
        """
        companies = self.data_fundamental.to_dict(orient="records")
        model_companies: List[IDataProviderCompany] = [
            IDataProviderCompany.model_validate(company) for company in companies
        ]
        model_companies = [
            target for target in model_companies if target.company_id in company_ids
        ]
        return model_companies

    def get_sbti_targets(self, companies: list) -> list:
        """
        For each of the companies, get the status of their target (Target set, Committed or No target) as it's known to
        the SBTi.
        
        :param companies: A list of companies. Each company should be a dict with a "company_name" and "company_id"
                            field.
        :return: The original list, enriched with a field called "sbti_target_status"
        """
        return self.data_fundamental[
            (
                self.data_fundamental["company_id"].isin(
                    [company["company_id"] for company in companies]
                )
                & self.data_fundamental["company_id"].notnull()
            )
        ].copy()


from typing import List
from pydantic import ValidationError
import logging

import pandas as pd
from ITR.data.data_provider import DataProvider
from ITR.interfaces import IDataProviderCompany, IDataProviderTarget


class CSVProvider(DataProvider):
    """
    Data provider skeleton for CSV files. This class serves primarily for testing purposes only!

    :param config: A dictionary containing a "path" field that leads to the path of the CSV file
    """

    def __init__(self, path: str, path_targets: str, encoding: str = "utf-8"):
        super().__init__()
        self.data = pd.read_csv(path, encoding=encoding)
        self.data_targets = pd.read_csv(path_targets, encoding=encoding)

    def get_targets(self, company_ids: list) -> List[IDataProviderTarget]:
        """
        Get all relevant targets for a list of company ids (ISIN). This method should return a list of
        IDataProviderTarget instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the targets
        """
        model_targets = self._target_df_to_model(self.data_targets)
        model_targets = [
            target for target in model_targets if target.company_id in company_ids
        ]
        return model_targets

    def _target_df_to_model(self, df_targets):
        """
        transforms target Dataframe into list of IDataProviderTarget instances

        :param df_targets: pandas Dataframe with targets
        :return: A list containing the targets
        """
        logger = logging.getLogger(__name__)
        targets = df_targets.to_dict(orient="records")
        model_targets: List[IDataProviderTarget] = []
        for target in targets:
            try:
                model_targets.append(IDataProviderTarget.model_validate(target))
            except ValidationError as e:
                logger.warning(
                    "(one of) the target(s) of company %s is invalid and will be skipped"
                    % target[self.c.COMPANY_NAME]
                )
                pass
        return model_targets

    def get_company_data(self, company_ids: list) -> List[IDataProviderCompany]:
        """
        Get all relevant data for a list of company ids (ISIN). This method should return a list of IDataProviderCompany
        instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the company data
        """
        companies = self.data.to_dict(orient="records")
        model_companies: List[IDataProviderCompany] = [
            IDataProviderCompany.model_validate(company) for company in companies
        ]
        model_companies = [
            target for target in model_companies if target.company_id in company_ids
        ]
        return model_companies

    def get_sbti_targets(self, companies: list) -> list:
        """
        For each of the companies, get the status of their target (Target set, Committed or No target) as it's known to
        the SBTi.

        :param companies: A list of companies. Each company should be a dict with a "company_name" and "company_id"
                            field.
        :return: The original list, enriched with a field called "sbti_target_status"
        """
        return self.data[
            (
                self.data["company_id"].isin(
                    [company["company_id"] for company in companies]
                )
                & self.data["company_id"].notnull()
            )
        ].copy()



from typing import Optional, List

import requests

from ITR.data.data_provider import DataProvider
from ITR.interfaces import IDataProviderTarget, IDataProviderCompany


class Bloomberg(DataProvider):
    """
    Data provider skeleton for Bloomberg.
    """

    def _request(self, endpoint: str, data: dict) -> Optional[object]:
        """
        Request data from the server.
        Note: This request does in no way reflect the actual implementation, this is only a stub to show what a
        potential API request COULD look like.

        :param endpoint: The endpoint of the API
        :param data: The data to send as a body
        :return: The returned data, None in case of an error.
        """
        try:
            headers = {"Authorization": "Basic: {}:{}".format("username", "password")}
            r = requests.post(
                "{}{}".format("host", endpoint), json=data, headers=headers
            )
            if r.status_code == 200:
                return r.json()
        except Exception as e:
            return None
        return None

    def get_targets(self, company_ids: List[str]) -> List[IDataProviderTarget]:
        """
        Get all relevant targets for a list of company ids (ISIN). This method should return a list of
        IDataProviderTarget instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the targets
        """
        # TODO: Make an API request
        # TODO: Transform the result into a dataframe
        # TODO: Make sure the columns align with those defined in the docstring
        raise NotImplementedError

    def get_company_data(self, company_ids: List[str]) -> List[IDataProviderCompany]:
        """
        Get all relevant data for a list of company ids (ISIN). This method should return a list of IDataProviderCompany
        instances.

        :param company_ids: A list of company IDs (ISINs)
        :return: A list containing the company data
        """
        # TODO: Make an API request
        # TODO: Transform the result into a dataframe
        # TODO: Make sure the columns align with those defined in the docstring
        raise NotImplementedError

    def get_sbti_targets(self, companies: list) -> list:
        """
        For each of the companies, get the status of their target (Target set, Committed or No target) as it's known to
        the SBTi.

        :param companies: A list of companies. Each company should be a dict with a "company_name" and "company_id"
                            field.
        :return: The original list, enriched with a field called "sbti_target_status"
        """
        # TODO: Make an API request
        # TODO: Extract the SBTi target status from the response
        # TODO: Enrich the original list with this data
        raise NotImplementedError
